import subprocess
import sys
import os

import dbus
import dbus.service
import dbus.glib

from landscape.ui.lib.polkit import PolicyKitMechanism, POLICY_NAME


SERVICE_NAME = "com.canonical.LandscapeClientRegistration"
INTERFACE_NAME = \
    "com.canonical.LandscapeClientRegistration.RegistrationInterface"
OBJECT_PATH = \
    "/com/canonical/LandscapeClientRegistration/RegistrationInterface"


class PermissionDeniedByPolicy(dbus.DBusException):
    _dbus_error_name = \
        "com.canonical.LandscapeClientRegistration.PermissionDeniedByPolicy"


class RegistrationError(dbus.DBusException):
    _dbus_error_name = \
        "com.canonical.LandscapeClientRegistration.RegistrationError"


class RegistrationMechanism(PolicyKitMechanism):
    """
    L{RegistrationMechanism} is a mechanism for invoking and observing client
    registration over DBus.  It utilises PolicyKit to ensure that only
    administrative users may use it.
    """

    def __init__(self, bus_name, bypass=False, conn=None):
        super(RegistrationMechanism, self).__init__(
            OBJECT_PATH, bus_name, PermissionDeniedByPolicy,
            bypass=bypass, conn=conn)
        self.process = None
        self.message_queue = []
        self.error_queue = []

    def _do_registration(self, config_path):
        self.register_notify("Trying to register ...\n")
        cmd = ["landscape-config", "--silent", "-c",
               os.path.abspath(config_path)]
        try:
            message = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
            self.register_notify(message)
            return True, message
        except subprocess.CalledProcessError, error:
            wait_phrase = "Please wait... "
            wait_phrase_index = error.output.find(wait_phrase)
            if wait_phrase_index > -1:
                message = error.output[wait_phrase_index + len(wait_phrase):]
            else:
                message = "Landscape configuration failed.\n%s" % error.output
            self.register_error(message)
            return False, message

    @dbus.service.signal(dbus_interface=INTERFACE_NAME,
                         signature='s')
    def register_notify(self, message):
        """
        L{register_notify} is a signal sent to subscribers.  It is not
        necessary for any actual work to occur in the method as it is called
        for the effect of invoking its decorator.
        """

    @dbus.service.signal(dbus_interface=INTERFACE_NAME,
                         signature='s')
    def register_error(self, message):
        """
        L{register_error} is a signal sent to subscribers.  It is not
        necessary for any actual work to occur in the method as it is called
        for the effect of invoking its decorator.
        """

    @dbus.service.signal(dbus_interface=INTERFACE_NAME,
                         signature='s')
    def register_succeed(self, message):
        """
        L{register_succeed} is a signal sent to subscribers.  It is not
        necessary for any actual work to occur in the method as it is called
        for the effect of invoking its decorator.
        """

    @dbus.service.signal(dbus_interface=INTERFACE_NAME,
                         signature='s')
    def register_fail(self, message):
        """
        L{register_fail} is a signal sent to subscribers.  It is not
        necessary for any actual work to occur in the method as it is called
        for the effect of invoking its decorator.
        """

    @dbus.service.method(INTERFACE_NAME,
                         in_signature="",
                         out_signature="b",
                         sender_keyword="sender",
                         connection_keyword="conn")
    def challenge(self, sender=None, conn=None):
        """
        Safely check if we can escalate permissions.
        """
        try:
            return self._is_allowed_by_policy(sender, conn, POLICY_NAME)
        except PermissionDeniedByPolicy:
            return False

    @dbus.service.method(INTERFACE_NAME,
                         in_signature="s",
                         out_signature="(bs)",
                         sender_keyword="sender",
                         connection_keyword="conn")
    def register(self, config_path, sender=None, conn=None):
        if self._is_allowed_by_policy(sender, conn, POLICY_NAME):
            succeed, message = self._do_registration(config_path)
            if succeed:
                message = "Registration message sent to Landscape server.\n"
                self.register_succeed(message)
                return (True, message)
            else:
                self.register_fail(message)
                return (False, message)

    def _do_disabling(self):
        cmd = ["landscape-config", "--disable"]
        try:
            subprocess.check_output(cmd, stderr=subprocess.STDOUT)
            return True
        except subprocess.CalledProcessError:
            return False

    @dbus.service.signal(dbus_interface=INTERFACE_NAME,
                         signature='')
    def disable_succeed(self):
        """
        L{disable_succeed} is a signal sent to subscribers.  It is not
        necessary for any actual work to occur in the method as it is called
        for the effect of invoking its decorator.
        """

    @dbus.service.signal(dbus_interface=INTERFACE_NAME,
                         signature='')
    def disable_fail(self):
        """
        L{disable_fail} is a signal sent to subscribers.  It is not
        necessary for any actual work to occur in the method as it is called
        for the effect of invoking its decorator.
        """

    @dbus.service.method(INTERFACE_NAME,
                         in_signature="",
                         out_signature="b",
                         sender_keyword="sender",
                         connection_keyword="conn")
    def disable(self, sender=None, conn=None):
        if self._is_allowed_by_policy(sender, conn, POLICY_NAME):
            if self._do_disabling():
                self.disable_succeed()
                return True
            else:
                self.disable_fail()
                return False

    @dbus.service.method(INTERFACE_NAME,
                         in_signature="",
                         out_signature="",
                         sender_keyword="sender",
                         connection_keyword="conn")
    def exit(self, sender=None, conn=None):
        """
        Exit this process.
        """
        sys.exit(0)



"""
A monitor plugin that collects data on a machine's network devices.
"""

from landscape.monitor.plugin import DataWatcher
from landscape.lib.network import get_active_device_info


class NetworkDevice(DataWatcher):

    message_type = "network-device"
    message_key = "devices"
    persist_name = message_type
    scope = "network"

    def __init__(self, device_info=get_active_device_info):
        super(NetworkDevice, self).__init__()
        self._device_info = device_info

    def register(self, registry):
        super(NetworkDevice, self).register(registry)
        self.call_on_accepted(self.message_type, self.exchange, True)

    def get_message(self):
        device_data = self._device_info()
        # Persist if the info is new.
        if self._persist.get("network-device-data") != device_data:
            self._persist.set("network-device-data", device_data)
            # We need to split the message in two top-level keys (see bug)
            device_speeds = []
            for device in device_data:
                speed_entry = {"interface": device["interface"]}
                speed_entry["speed"] = device.pop("speed")
                speed_entry["duplex"] = device.pop("duplex")
                device_speeds.append(speed_entry)

            return {"type": self.message_type,
                    "devices": device_data,
                    "device-speeds": device_speeds}



from landscape.schema import (
    Message, KeyDict, Dict, List, Tuple,
    Bool, Int, Float, Bytes, Unicode, Constant, Any)

# When adding a new schema, which deprecates an older schema, the recommended
# naming convention, is to name it SCHEMA_NAME_ and the last API version that
# the schema works with.
#
# i.e. if I have USERS and I'm deprecating it, in API 2.2, then USERS becomes
# USERS_2_1


process_info = KeyDict({"pid": Int(),
                        "name": Unicode(),
                        "state": Bytes(),
                        "sleep-average": Int(),
                        "uid": Int(),
                        "gid": Int(),
                        "vm-size": Int(),
                        "start-time": Int(),
                        "percent-cpu": Float()},
                       # Optional for backwards compatibility
                       optional=["vm-size", "sleep-average", "percent-cpu"])

ACTIVE_PROCESS_INFO = Message(
    "active-process-info",
    {"kill-processes": List(Int()),
     "kill-all-processes": Bool(),
     "add-processes": List(process_info),
     "update-processes": List(process_info)},
    # XXX Really we don't want all three of these keys to be optional:
    # we always want _something_...
    optional=["add-processes", "update-processes", "kill-processes",
              "kill-all-processes"])

COMPUTER_UPTIME = Message(
    "computer-uptime",
    {"startup-times": List(Int()),
     "shutdown-times": List(Int())},
    # XXX Again, one or the other.
    optional=["startup-times", "shutdown-times"])

CLIENT_UPTIME = Message(
    "client-uptime",
    {"period": Tuple(Float(), Float()),
     "components": List(Int())},
    optional=["components"])  # just for backwards compatibility

OPERATION_RESULT = Message(
    "operation-result",
    {"operation-id": Int(),
     "status": Int(),
     "result-code": Int(),
     "result-text": Unicode()},
    optional=["result-code", "result-text"])

#ACTION_INFO is obsolete.
ACTION_INFO = Message(
    "action-info",
    {"response-id": Int(),
     "success": Bool(),
     "kind": Bytes(),
     "parameters": Bytes()})

COMPUTER_INFO = Message(
    "computer-info",
    {"hostname": Unicode(),
     "total-memory": Int(),
     "total-swap": Int(),
     "annotations": Dict(Unicode(), Unicode())},
    # Not sure why these are all optional, but it's explicitly tested
    # in the server
    optional=["hostname", "total-memory", "total-swap", "annotations"])

DISTRIBUTION_INFO = Message(
    "distribution-info",
    {"distributor-id": Unicode(),
     "description": Unicode(),
     "release": Unicode(),
     "code-name": Unicode()},
    # all optional because the lsb-release file may not have all data.
    optional=["distributor-id", "description", "release", "code-name"])

CLOUD_METADATA = Message(
    "cloud-instance-metadata",
    {"instance-id": Unicode(),
     "ami-id": Unicode(),
     "instance-type": Unicode()})


hal_data = Dict(Unicode(),
                Any(Unicode(), List(Unicode()), Bool(), Int(), Float()))

HARDWARE_INVENTORY = Message("hardware-inventory", {
    "devices": List(Any(Tuple(Constant("create"), hal_data),
                        Tuple(Constant("update"),
                              Unicode(),  # udi,
                              hal_data,  # creates,
                              hal_data,  # updates,
                              hal_data),  # deletes
                        Tuple(Constant("delete"),
                              Unicode()),
                        ),
                    )})


HARDWARE_INFO = Message("hardware-info", {
    "data": Unicode()})

juju_data = {"environment-uuid": Unicode(),
             "api-addresses": List(Unicode()),
             "unit-name": Unicode()}

# The copy is needed because Message mutates the dictionary
JUJU_INFO = Message("juju-info", juju_data.copy())

LOAD_AVERAGE = Message("load-average", {
    "load-averages": List(Tuple(Int(), Float())),
    })

CPU_USAGE = Message("cpu-usage", {
    "cpu-usages": List(Tuple(Int(), Float())),
    })

CEPH_USAGE = Message("ceph-usage", {
    "ceph-usages": List(Tuple(Int(), Float())),
    "ring-id": Unicode(),
    })

SWIFT_DEVICE_INFO = Message("swift-device-info", {
    "swift-device-info": List(
        KeyDict({"device": Unicode(), "mounted": Bool()}))
    })

KEYSTONE_TOKEN = Message("keystone-token", {
    "data": Any(Bytes(), Constant(None))
})

CHANGE_HA_SERVICE = Message(
    "change-ha-service",
    {"service-name": Bytes(),  # keystone
     "unit-name": Bytes(),     # keystone-9
     "state": Bytes()})        # online or standby

MEMORY_INFO = Message("memory-info", {
    "memory-info": List(Tuple(Float(), Int(), Int())),
    })

RESYNCHRONIZE = Message(
    "resynchronize",
    {"operation-id": Int()},
    # operation-id is only there if it's a response to a server-initiated
    # resynchronize.
    optional=["operation-id"])

MOUNT_ACTIVITY = Message("mount-activity", {
    "activities": List(Tuple(Float(), Unicode(), Bool()))})


MOUNT_INFO = Message("mount-info", {
    "mount-info": List(Tuple(Float(),
                             KeyDict({"mount-point": Unicode(),
                                      "device": Unicode(),
                                      "filesystem": Unicode(),
                                      "total-space": Int()})
                             )),
    })

FREE_SPACE = Message("free-space", {
    "free-space": List(Tuple(Float(), Unicode(), Int()))})


REGISTER = Message(
    "register",
    # The term used in the UI is actually 'registration_key', but we keep
    # the message schema field as 'registration_password' in case a new
    # client contacts an older server.
    {"registration_password": Any(Unicode(), Constant(None)),
     "computer_title": Unicode(),
     "hostname": Unicode(),
     "account_name": Unicode(),
     "tags": Any(Unicode(), Constant(None)),
     "vm-info": Bytes(),
     "container-info": Unicode(),
     "juju-info": KeyDict(juju_data),
     "access_group": Unicode()},
    optional=["registration_password", "hostname", "tags", "vm-info",
              "container-info", "juju-info", "unicode", "access_group"])


REGISTER_PROVISIONED_MACHINE = Message(
    "register-provisioned-machine",
    {"otp": Bytes()})


REGISTER_CLOUD_VM = Message(
    "register-cloud-vm",
    {"hostname": Unicode(),
     "otp": Any(Bytes(), Constant(None)),
     "instance_key": Unicode(),
     "account_name": Any(Unicode(), Constant(None)),
     "registration_password": Any(Unicode(), Constant(None)),
     "reservation_key": Unicode(),
     "public_hostname": Unicode(),
     "local_hostname": Unicode(),
     "kernel_key": Any(Unicode(), Constant(None)),
     "ramdisk_key": Any(Unicode(), Constant(None)),
     "launch_index": Int(),
     "image_key": Unicode(),
     "tags": Any(Unicode(), Constant(None)),
     "vm-info": Bytes(),
     "public_ipv4": Unicode(),
     "local_ipv4": Unicode(),
     "access_group": Unicode()},
    optional=["tags", "vm-info", "public_ipv4", "local_ipv4", "access_group"])

TEMPERATURE = Message("temperature", {
    "thermal-zone": Unicode(),
    "temperatures": List(Tuple(Int(), Float())),
    })

PROCESSOR_INFO = Message(
    "processor-info",
    {"processors": List(KeyDict({"processor-id": Int(),
                                 "vendor": Unicode(),
                                 "model": Unicode(),
                                 "cache-size": Int(),
                                 },
                                optional=["vendor", "cache-size"]))})

user_data = KeyDict({
    "uid": Int(),
    "username": Unicode(),
    "name": Any(Unicode(), Constant(None)),
    "enabled": Bool(),
    "location": Any(Unicode(), Constant(None)),
    "home-phone": Any(Unicode(), Constant(None)),
    "work-phone": Any(Unicode(), Constant(None)),
    "primary-gid": Any(Int(), Constant(None)),
    "primary-groupname": Unicode()},
    optional=["primary-groupname", "primary-gid"])

group_data = KeyDict({
    "gid": Int(),
    "name": Unicode()})

USERS = Message(
    "users",
    {"operation-id": Int(),
     "create-users": List(user_data),
     "update-users": List(user_data),
     "delete-users": List(Unicode()),
     "create-groups": List(group_data),
     "update-groups": List(group_data),
     "delete-groups": List(Unicode()),
     "create-group-members": Dict(Unicode(), List(Unicode())),
     "delete-group-members": Dict(Unicode(), List(Unicode())),
     },
    # operation-id is only there for responses, and all other are
    # optional as long as one of them is there (no way to say that yet)
    optional=["operation-id", "create-users", "update-users", "delete-users",
              "create-groups", "update-groups", "delete-groups",
              "create-group-members", "delete-group-members"])

USERS_2_1 = Message(
    "users",
    {"operation-id": Int(),
     "create-users": List(user_data),
     "update-users": List(user_data),
     "delete-users": List(Int()),
     "create-groups": List(group_data),
     "update-groups": List(group_data),
     "delete-groups": List(Int()),
     "create-group-members": Dict(Int(), List(Int())),
     "delete-group-members": Dict(Int(), List(Int())),
     },
    # operation-id is only there for responses, and all other are
    # optional as long as one of them is there (no way to say that yet)
    optional=["operation-id", "create-users", "update-users", "delete-users",
              "create-groups", "update-groups", "delete-groups",
              "create-group-members", "delete-group-members"])

USERS_2_0 = Message(
    "users",
    {"operation-id": Int(),
     "create-users": List(user_data),
     "update-users": List(user_data),
     "delete-users": List(Int()),
     "create-groups": List(group_data),
     "update-groups": List(group_data),
     "delete-groups": List(Int()),
     "create-group-members": Dict(Int(), List(Int())),
     "delete-group-members": Dict(Int(), List(Int())),
     },
    # operation-id is only there for responses, and all other are
    # optional as long as one of them is there (no way to say that yet)
    optional=["operation-id", "create-users", "update-users", "delete-users",
              "create-groups", "update-groups", "delete-groups",
              "create-group-members", "delete-group-members"])

opt_str = Any(Unicode(), Constant(None))
OLD_USERS = Message(
    "users",
    {"users": List(KeyDict({"username": Unicode(),
                            "uid": Int(),
                            "realname": opt_str,
                            "location": opt_str,
                            "home-phone": opt_str,
                            "work-phone": opt_str,
                            "enabled": Bool()},
                           optional=["location", "home-phone", "work-phone"])),
     "groups": List(KeyDict({"gid": Int(),
                             "name": Unicode(),
                             "members": List(Unicode())}))},
    optional=["groups"])

package_ids_or_ranges = List(Any(Tuple(Int(), Int()), Int()))
PACKAGES = Message(
    "packages",
    {"installed": package_ids_or_ranges,
     "available": package_ids_or_ranges,
     "available-upgrades": package_ids_or_ranges,
     "locked": package_ids_or_ranges,
     "not-installed": package_ids_or_ranges,
     "not-available": package_ids_or_ranges,
     "not-available-upgrades": package_ids_or_ranges,
     "not-locked": package_ids_or_ranges},
    optional=["installed", "available", "available-upgrades", "locked",
              "not-available", "not-installed", "not-available-upgrades",
              "not-locked"])

package_locks = List(Tuple(Unicode(), Unicode(), Unicode()))
PACKAGE_LOCKS = Message(
    "package-locks",
    {"created": package_locks,
     "deleted": package_locks},
    optional=["created", "deleted"])

CHANGE_PACKAGE_HOLDS = Message(
    "change-package-holds",
    {"created": List(Unicode()),
     "deleted": List(Unicode())},
    optional=["created", "deleted"])

CHANGE_PACKAGES_RESULT = Message(
    "change-packages-result",
    {"operation-id": Int(),
     "must-install": List(Any(Int(), Constant(None))),
     "must-remove": List(Any(Int(), Constant(None))),
     "result-code": Int(),
     "result-text": Unicode()},
    optional=["result-text", "must-install", "must-remove"])

UNKNOWN_PACKAGE_HASHES = Message("unknown-package-hashes", {
    "hashes": List(Bytes()),
    "request-id": Int(),
    })

PACKAGE_REPORTER_RESULT = Message("package-reporter-result", {
    "code": Int(),
    "err": Unicode()})

ADD_PACKAGES = Message("add-packages", {
    "packages": List(KeyDict({"name": Unicode(),
                              "description": Unicode(),
                              "section": Unicode(),
                              "relations": List(Tuple(Int(), Unicode())),
                              "summary": Unicode(),
                              "installed-size": Any(Int(), Constant(None)),
                              "size": Any(Int(), Constant(None)),
                              "version": Unicode(),
                              "type": Int(),
                              })),
    "request-id": Int(),
    })

TEXT_MESSAGE = Message("text-message", {
    "message": Unicode()})

TEST = Message(
    "test",
    {"greeting": Bytes(),
     "consistency-error": Bool(),
     "echo": Bytes(),
     "sequence": Int()},
    optional=["greeting", "consistency-error", "echo", "sequence"])

# The tuples are timestamp, value
GRAPH_DATA = KeyDict({"values": List(Tuple(Float(), Float())),
                      "error": Unicode(),
                      "script-hash": Bytes()})

CUSTOM_GRAPH = Message("custom-graph", {
    "data": Dict(Int(), GRAPH_DATA)})

# XXX This is kept for backward compatibility, it can eventually be removed
# when all clients will support REBOOT_REQUIRED_INFO
REBOOT_REQUIRED = Message(
    "reboot-required",
    {"flag": Bool()})

REBOOT_REQUIRED_INFO = Message(
    "reboot-required-info",
    {"flag": Bool(),
     "packages": List(Unicode())},
    optional=["flag", "packages"])

APT_PREFERENCES = Message(
    "apt-preferences",
    {"data": Any(Dict(Unicode(), Unicode()), Constant(None))})

EUCALYPTUS_INFO = Message(
    "eucalyptus-info",
    {"basic_info": Dict(Bytes(), Any(Bytes(), Constant(None))),
     "walrus_info": Bytes(),
     "cluster_controller_info": Bytes(),
     "storage_controller_info": Bytes(),
     "node_controller_info": Bytes(),
     "capacity_info": Bytes()},
    optional=["capacity_info"])

EUCALYPTUS_INFO_ERROR = Message(
    "eucalyptus-info-error",
    {"error": Bytes()})

# The network-device message is split in two top level keys because we don't
# support adding sub-keys in a backwards-compatible way (only top-level keys).
# New servers will see an optional device-speeds key, and old servers will
# simply ignore the extra info..
NETWORK_DEVICE = Message(
    "network-device",
    {"devices": List(KeyDict({"interface": Bytes(),
                              "ip_address": Bytes(),
                              "mac_address": Bytes(),
                              "broadcast_address": Bytes(),
                              "netmask": Bytes(),
                              "flags": Int()})),

     "device-speeds": List(KeyDict({"interface": Bytes(),
                                    "speed": Int(),
                                    "duplex": Bool()}))},
    optional=["device-speeds"])


NETWORK_ACTIVITY = Message(
    "network-activity",
    # Dict maps interfaces to their network activity. The network activity of
    # an interface a is a list of 3-tuples (step, in, out), where 'step' is the
    # time interval and 'in'/'out' are number of bytes received/sent over the
    # interval.
    {"activities": Dict(Bytes(), List(Tuple(Int(), Int(), Int())))})

UPDATE_MANAGER_INFO = Message("update-manager-info", {"prompt": Unicode()})


message_schemas = {}
for schema in [ACTIVE_PROCESS_INFO, COMPUTER_UPTIME, CLIENT_UPTIME,
               OPERATION_RESULT, COMPUTER_INFO, DISTRIBUTION_INFO,
               HARDWARE_INVENTORY, HARDWARE_INFO, LOAD_AVERAGE, MEMORY_INFO,
               RESYNCHRONIZE, MOUNT_ACTIVITY, MOUNT_INFO, FREE_SPACE,
               REGISTER, REGISTER_CLOUD_VM, REGISTER_PROVISIONED_MACHINE,
               TEMPERATURE, PROCESSOR_INFO, USERS, PACKAGES, PACKAGE_LOCKS,
               CHANGE_PACKAGES_RESULT, UNKNOWN_PACKAGE_HASHES,
               ADD_PACKAGES, PACKAGE_REPORTER_RESULT, TEXT_MESSAGE, TEST,
               CUSTOM_GRAPH, REBOOT_REQUIRED, APT_PREFERENCES, EUCALYPTUS_INFO,
               EUCALYPTUS_INFO_ERROR, NETWORK_DEVICE, NETWORK_ACTIVITY,
               REBOOT_REQUIRED_INFO, UPDATE_MANAGER_INFO, CPU_USAGE,
               CEPH_USAGE, SWIFT_DEVICE_INFO, KEYSTONE_TOKEN,
               CHANGE_HA_SERVICE, JUJU_INFO, CLOUD_METADATA]:
    message_schemas[schema.type] = schema



from landscape.monitor.cpuusage import CPUUsage, LAST_MESURE_KEY
from landscape.tests.helpers import LandscapeTest, MonitorHelper


class CPUUsagePluginTest(LandscapeTest):

    helpers = [MonitorHelper]

    def _write_stat_file(self, contents):
        statfile = self.makeFile()
        with open(statfile, "w") as f:
            f.write(contents)
        return statfile

    def test_get_cpu_usage_file_unreadable(self):
        """
        When the file is unreadable or somehow creates an IOError (like when
        it doesn't exist), the method returns None.
        """
        self.log_helper.ignore_errors("Could not open.*")
        thefile = "/tmp/whatever/I/do/not/exist"
        plugin = CPUUsage(create_time=self.reactor.time)
        self.monitor.add(plugin)

        result = plugin._get_cpu_usage(stat_file=thefile)
        self.assertIs(None, result)

    def test_get_cpu_usage_file_not_changed(self):
        """
        When the stat file did not change between calls, the
        C{_get_cpu_usage} method returns None.
        """
        contents1 = "cpu  100 100 100 100 100 100 100 0 0 0"

        thefile = self._write_stat_file(contents1)
        plugin = CPUUsage(create_time=self.reactor.time)
        self.monitor.add(plugin)

        result = plugin._get_cpu_usage(stat_file=thefile)
        # The first run will return None since we don't have a previous measure
        # yet.
        self.assertIs(None, result)

        result = plugin._get_cpu_usage(stat_file=thefile)
        self.assertIs(None, result)

    def test_get_cpu_usage_multiline_files(self):
        """
        The C{_get_cpu_usage} method parses multiline stat files correctly.
        """
        contents1 = "cpu  100 100 100 100 100 100 100 0 0 0\nsome garbage"

        thefile = self._write_stat_file(contents1)
        plugin = CPUUsage(create_time=self.reactor.time)
        self.monitor.add(plugin)

        result = plugin._get_cpu_usage(stat_file=thefile)
        # The first run will return None since we don't have a previous measure
        # yet.
        self.assertIs(None, result)

        result = plugin._get_cpu_usage(stat_file=thefile)
        self.assertIs(None, result)

    def test_get_cpu_usage_100_percent_usage(self):
        """
        When two consecutive calls to C{_get_cpu_usage} show a CPU usage of
        100%, the method returns 1.
        """
        contents1 = "cpu  100 100 100 100 100 100 100 0 0 0"
        contents2 = "cpu  200 100 100 100 100 100 100 0 0 0"

        thefile = self._write_stat_file(contents1)
        thefile2 = self._write_stat_file(contents2)
        plugin = CPUUsage(create_time=self.reactor.time)
        self.monitor.add(plugin)

        result = plugin._get_cpu_usage(stat_file=thefile)
        # The first run will return None since we don't have a previous measure
        # yet.
        self.assertIs(None, result)

        result = plugin._get_cpu_usage(stat_file=thefile2)
        self.assertEqual(result, 1)

    def test_get_cpu_usage_0_percent_usage(self):
        """
        When two consecutive calls to C{_get_cpu_usage} show a CPU usage of
        0% (all the changes are in the idle column) the method returns 0.
        """
        contents1 = "cpu  100 100 100 100 100 100 100 0 0 0"
        contents2 = "cpu  100 100 100 200 100 100 100 0 0 0"

        thefile = self._write_stat_file(contents1)
        thefile2 = self._write_stat_file(contents2)
        plugin = CPUUsage(create_time=self.reactor.time)
        self.monitor.add(plugin)

        result = plugin._get_cpu_usage(stat_file=thefile)
        # The first run will return None since we don't have a previous measure
        # yet.
        self.assertIs(None, result)

        result = plugin._get_cpu_usage(stat_file=thefile2)
        self.assertEqual(result, 0)

    def test_get_cpu_usage_50_percent_usage(self):
        """
        When two consecutive calls to C{_get_cpu_usage} show a CPU usage of
        50% (as much changed in an "active" column that in the idle column)
        the method returns 0.5.
        """
        contents1 = "cpu  100 100 100 100 100 100 100 0 0 0"
        contents2 = "cpu  200 100 100 200 100 100 100 0 0 0"

        thefile = self._write_stat_file(contents1)
        thefile2 = self._write_stat_file(contents2)
        plugin = CPUUsage(create_time=self.reactor.time)
        self.monitor.add(plugin)

        result = plugin._get_cpu_usage(stat_file=thefile)
        # The first run will return None since we don't have a previous measure
        # yet.
        self.assertIs(None, result)

        result = plugin._get_cpu_usage(stat_file=thefile2)
        self.assertEqual(result, 0.5)

    def test_get_cpu_usage_after_reboot(self):
        """
        When the computer just rebooted, we might have a case where the
        previous values are larger that the current values (since the kernel
        counts quantums allocated since boot). In this case, the method should
        return None.
        """
        contents1 = "cpu  100 100 100 100 100 100 100 0 0 0"

        measure1 = (700, 100)
        measure2 = (900, 10)

        thefile = self._write_stat_file(contents1)
        plugin = CPUUsage(create_time=self.reactor.time)
        self.monitor.add(plugin)
        plugin._persist.set(LAST_MESURE_KEY, measure2)

        result = plugin._get_cpu_usage(stat_file=thefile)
        self.assertIs(None, result)
        self.assertEqual(measure1, plugin._persist.get(LAST_MESURE_KEY))

    def test_create_message(self):
        """
        Calling create_message returns an expected message.
        """
        plugin = CPUUsage(create_time=self.reactor.time)
        self.monitor.add(plugin)

        plugin._cpu_usage_points = []
        message = plugin.create_message()
        self.assertIn("type", message)
        self.assertEqual(message["type"], "cpu-usage")
        self.assertIn("cpu-usages", message)
        cpu_usages = message["cpu-usages"]
        self.assertEqual(len(cpu_usages), 0)

        point = (60, 1.0)
        plugin._cpu_usage_points = [point]
        message = plugin.create_message()
        self.assertIn("type", message)
        self.assertEqual(message["type"], "cpu-usage")
        self.assertIn("cpu-usages", message)
        cpu_usages = message["cpu-usages"]
        self.assertEqual(len(cpu_usages), 1)
        self.assertEqual(point, cpu_usages[0])

    def test_never_exchange_empty_messages(self):
        """
        The plugin will create a message with an empty
        C{cpu-usages} list when no previous data is available.  If an empty
        message is created during exchange, it should not be queued.
        """
        self.mstore.set_accepted_types(["cpu-usage"])

        plugin = CPUUsage(create_time=self.reactor.time)
        self.monitor.add(plugin)

        self.monitor.exchange()
        self.assertEqual(len(self.mstore.get_pending_messages()), 0)

    def test_exchange_messages(self):
        """
        The CPU usage plugin queues message when manager.exchange()
        is called.
        """
        self.mstore.set_accepted_types(["cpu-usage"])

        plugin = CPUUsage(create_time=self.reactor.time)
        plugin._cpu_usage_points = [(60, 1.0)]
        self.monitor.add(plugin)

        self.monitor.exchange()

        self.assertMessages(self.mstore.get_pending_messages(),
                            [{"type": "cpu-usage",
                              "cpu-usages": [(60, 1.0)]}])

    def test_no_message_if_not_accepted(self):
        """
        Don't add any messages at all if the broker isn't currently
        accepting their type.
        """
        interval = 30

        plugin = CPUUsage(create_time=self.reactor.time,
                          interval=interval)

        self.monitor.add(plugin)

        self.reactor.advance(self.monitor.step_size * 2)
        self.monitor.exchange()

        self.mstore.set_accepted_types(["cpu-usage"])
        self.assertMessages(list(self.mstore.get_pending_messages()), [])

    def test_plugin_run(self):
        """
        The plugin's run() method fills in the _cpu_usage_points with
        accumulated samples after each C{monitor.step_size} period.
        """
        plugin = CPUUsage(create_time=self.reactor.time)
        self.monitor.add(plugin)

        def fake_get_cpu_usage(self):
            return 1.0

        plugin._get_cpu_usage = fake_get_cpu_usage
        self.reactor.advance(self.monitor.step_size * 2)

        self.assertNotEqual([], plugin._cpu_usage_points)
        self.assertEqual([(300, 1.0), (600, 1.0)], plugin._cpu_usage_points)

    def test_plugin_run_with_None(self):
        """
        The plugin's run() method fills in the _cpu_usage_points with
        accumulated samples after each C{monitor.step_size} period.
        Holes in the data (in case of error the method returns None) are
        handled gracefully, and are filled with averaged data.
        """
        plugin = CPUUsage(create_time=self.reactor.time)
        self.monitor.add(plugin)

        def fake_get_cpu_usage(self):
            return 1.0

        def fake_get_cpu_usage_none(self):
            return None

        plugin._get_cpu_usage = fake_get_cpu_usage
        self.reactor.advance(self.monitor.step_size)
        plugin._get_cpu_usage = fake_get_cpu_usage_none
        self.reactor.advance(self.monitor.step_size)

        self.assertNotEqual([], plugin._cpu_usage_points)
        self.assertEqual([(300, 1.0)], plugin._cpu_usage_points)

        # If we record values once again the "blank" period will be smoothed
        # over with the new points.
        plugin._get_cpu_usage = fake_get_cpu_usage
        self.reactor.advance(self.monitor.step_size)
        self.assertEqual([(300, 1.0), (600, 1.0), (900, 1.0)],
                         plugin._cpu_usage_points)



from landscape.upgraders import broker, monitor, package


UPGRADE_MANAGERS = {
    # these should not be hardcoded
    "broker": broker.upgrade_manager,
    "monitor": monitor.upgrade_manager,
    "package": package.upgrade_manager}



from landscape.sysinfo.sysinfo import SysInfoPluginRegistry
from landscape.sysinfo.load import Load
from landscape.tests.helpers import LandscapeTest


class LoadTest(LandscapeTest):

    def setUp(self):
        super(LoadTest, self).setUp()
        self.load = Load()
        self.sysinfo = SysInfoPluginRegistry()
        self.sysinfo.add(self.load)

    def test_run_returns_succeeded_deferred(self):
        self.assertIs(None, self.successResultOf(self.load.run()))

    def test_run_adds_header(self):
        mock = self.mocker.replace("os.getloadavg")
        mock()
        self.mocker.result((1.5, 0, 0))
        self.mocker.replay()

        self.load.run()

        self.assertEqual(self.sysinfo.get_headers(),
                         [("System load", "1.5")])



from logging import info, exception

from twisted.internet.defer import maybeDeferred

from landscape.log import format_object
from landscape.lib.twisted_util import gather_results
from landscape.amp import remote


class HandlerNotFoundError(Exception):
    """A handler for the given message type was not found."""


class BrokerClientPlugin(object):
    """A convenience for writing L{BrokerClient} plugins.

    This provides a register method which will set up a bunch of
    reactor handlers in the idiomatic way.

    If C{run} is defined on subclasses, it will be called every C{run_interval}
    +seconds after being registered.

    @cvar run_interval: The interval, in seconds, to execute the C{run} method.
        If set to C{None}, then C{run} will not be scheduled.
    @cvar run_immediately: If C{True} the plugin will be run immediately after
        it is registered.
    @ivar _session_id: the session id to be passed when sending messages via
        the broker.  This variable is set by the C{register} method and
        should only need to be renewed when a re-synchronisation request is
        sent. See L{landscape.broker.server.BrokerServer.send_message} for
        more details.
    """
    run_interval = 5
    run_immediately = False
    scope = None  # Global scope
    _session_id = None
    _loop = None

    def register(self, client):
        self.client = client
        self.client.reactor.call_on("resynchronize", self._resynchronize)
        deferred = self.client.broker.get_session_id(scope=self.scope)
        deferred.addCallback(self._got_session_id)

    @property
    def registry(self):
        """An alias for the C{client} attribute."""
        return self.client

    def call_on_accepted(self, type, callable, *args, **kwargs):
        """
        Register a callback fired upon a C{message-type-acceptance-changed}.
        """

        def acceptance_changed(acceptance):
            if acceptance:
                return callable(*args, **kwargs)

        self.client.reactor.call_on(("message-type-acceptance-changed", type),
                                    acceptance_changed)

    def _resynchronize(self, scopes=None):
        """
        Handle the 'resynchronize' event.  Subclasses should do any clear-down
        operations specific to their state within an implementation of the
        L{_reset} method.
        """
        if not (scopes is None or self.scope in scopes):
            # This resynchronize event is out of scope for us. Do nothing
            return

        # Because the broker will drop session IDs already associated to scope
        # of the resynchronisation, it isn't safe to send messages until the
        # client has received a new session ID.  Therefore we pause any calls
        # to L{run} by cancelling L{_loop}, this will be restarted in
        # L{_got_session_id}.
        if self._loop is not None:
            self.client.reactor.cancel_call(self._loop)

        # Do any state clean up required by the plugin.
        self._reset()

        deferred = self.client.broker.get_session_id(scope=self.scope)
        deferred.addCallback(self._got_session_id)
        return deferred

    def _reset(self):
        """
        Reset plugin specific state.

        Sub-classes should override this method to clear down data for
        resynchronisation.  Sub-classes with no state can simply ignore this.
        """

    def _got_session_id(self, session_id):
        """Save the session ID and invoke the C{run} method.

        We set the C{_session_id} attribute on the instance because it's
        required in order to send messages.  See
        L{BrokerService.get_session_id}.
        """
        self._session_id = session_id
        if getattr(self, "run", None) is not None:
            if self.run_immediately:
                self.run()
            if self.run_interval is not None:
                self._loop = self.client.reactor.call_every(self.run_interval,
                                                            self.run)


class BrokerClient(object):
    """Basic plugin registry for clients that have to deal with the broker.

    This knows about the needs of a client when dealing with the Landscape
    broker, including interest in messages of a particular type delivered
    by the broker to the client.

    @cvar name: The name used when registering to the broker, it must be
        defined by sub-classes.
    @ivar broker: A reference to a connected L{RemoteBroker}, it must be set
        by the connecting machinery at service startup.

    @param reactor: A L{LandscapeReactor}.
    """
    name = "client"

    def __init__(self, reactor):
        super(BrokerClient, self).__init__()
        self.reactor = reactor
        self.broker = None
        self._registered_messages = {}
        self._plugins = []
        self._plugin_names = {}

        # Register event handlers
        self.reactor.call_on("impending-exchange", self.notify_exchange)
        self.reactor.call_on("broker-reconnect", self.handle_reconnect)

    @remote
    def ping(self):
        """Return C{True}"""
        return True

    def add(self, plugin):
        """Add a plugin.

        The plugin's C{register} method will be called with this broker client
        as its argument.

        If the plugin has a C{plugin_name} attribute, it will be possible to
        look up the plugin later with L{get_plugin}.
        """
        info("Registering plugin %s.", format_object(plugin))
        self._plugins.append(plugin)
        if hasattr(plugin, 'plugin_name'):
            self._plugin_names[plugin.plugin_name] = plugin
        plugin.register(self)

    def get_plugins(self):
        """Get the list of plugins."""
        return self._plugins[:]

    def get_plugin(self, name):
        """Get a particular plugin by name."""
        return self._plugin_names[name]

    def register_message(self, type, handler):
        """
        Register interest in a particular type of Landscape server->client
        message.

        @param type: The type of message to register C{handler} for.
        @param handler: A callable taking a message as a parameter, called
            when messages of C{type} are received.
        @return: A C{Deferred} that will fire when registration completes.
        """
        self._registered_messages[type] = handler
        return self.broker.register_client_accepted_message_type(type)

    def dispatch_message(self, message):
        """Run the handler registered for the type of the given message.

        @return: The return value of the handler, if found.
        @raises: HandlerNotFoundError if the handler was not found
        """
        type = message["type"]
        handler = self._registered_messages.get(type)
        if handler is None:
            raise HandlerNotFoundError(type)
        try:
            return handler(message)
        except:
            exception("Error running message handler for type %r: %r"
                      % (type, handler))

    @remote
    def message(self, message):
        """Call C{dispatch_message} for the given C{message}.

        @return: A boolean indicating if a handler for the message was found.
        """
        try:
            self.dispatch_message(message)
            return True
        except HandlerNotFoundError:
            return False

    def exchange(self):
        """Call C{exchange} on all plugins."""
        for plugin in self.get_plugins():
            if hasattr(plugin, "exchange"):
                try:
                    plugin.exchange()
                except:
                    exception("Error during plugin exchange")

    def notify_exchange(self):
        """Notify all plugins about an impending exchange."""
        info("Got notification of impending exchange. Notifying all plugins.")
        self.exchange()

    @remote
    def fire_event(self, event_type, *args, **kwargs):
        """Fire an event of a given type.

        @return: A L{Deferred} resulting in a list of returns values of
            the fired event handlers, in the order they were fired.
        """
        if event_type == "message-type-acceptance-changed":
            message_type = args[0]
            acceptance = args[1]
            results = self.reactor.fire((event_type, message_type), acceptance)
        else:
            results = self.reactor.fire(event_type, *args, **kwargs)
        return gather_results([
            maybeDeferred(lambda x: x, result) for result in results])

    def handle_reconnect(self):
        """Called when the connection with the broker is established again.

        The following needs to be done:

          - Re-register any previously registered message types, so the broker
            knows we have interest on them.

          - Re-register ourselves as client, so the broker knows we exist and
            will talk to us firing events and dispatching messages.
        """
        for type in self._registered_messages:
            self.broker.register_client_accepted_message_type(type)
        self.broker.register_client(self.name)

    @remote
    def exit(self):
        """Stop the reactor and exit the process."""
        # Stop with a short delay to give a chance to reply to the caller when
        # this method is invoked over AMP (typically by the broker, see also
        # landscape.broker.server.BrokerServer.exit).
        self.reactor.call_later(0.1, self.reactor.stop)



"""
Support code for the C{landscape-message} utility, which sends a text
message to the Landscape web UI via the landscape-client's dbus
messaging service (see L{landscape.plugins.dbus_message}).
"""

import sys

from landscape.lib.log import log_failure
from landscape.reactor import LandscapeReactor
from landscape.broker.amp import RemoteBrokerConnector
from landscape.deployment import Configuration


class AcceptedTypeError(Exception):
    """
    Raised when a message is sent without 'text-message' being an
    accepted type.
    """


class EmptyMessageError(Exception):
    """Raised when an empty message is provied."""


def send_message(text, broker):
    """Add a message to the queue via a remote broker.

    The message is of type C{text-message}.

    @param broker: A connected L{RemoteBroker} object to use to send
        the message.
    @return: A L{Deferred} which will fire with the result of the send.
    """
    def got_session_id(session_id):
        response = broker.send_message(message, session_id, True)
        return response

    message = {"type": "text-message", "message": text}
    result = broker.get_session_id()
    result.addCallback(got_session_id)
    return result


def got_result(result):
    print u"Message sent."


def get_message(args):
    encoding = sys.stdin.encoding or "UTF-8"
    if len(args) < 2:
        print ("Please enter your message, and send EOF (Control + D after "
               "newline) when done.")
        message = sys.stdin.read().decode(encoding)
    else:
        message = u" ".join([x.decode(encoding) for x in args[1:]])
    if not message:
        raise EmptyMessageError("Text messages may not be empty.")
    return message


def got_accepted_types(accepted_types, broker, args):
    if not "text-message" in accepted_types:
        raise AcceptedTypeError("Text messages may not be created.  Is "
                                "Landscape Client registered with the server?")
    message = get_message(args)
    d = send_message(message, broker)
    d.addCallback(got_result)
    return d


def run(args=sys.argv):
    """Send a message to Landscape.

    This function runs a Twisted reactor, prints various status
    messages, and exits the process.
    """
    reactor = LandscapeReactor()
    config = Configuration()
    config.load(args)

    def got_connection(broker):
        result = broker.get_accepted_message_types()
        return result.addCallback(got_accepted_types, broker, args)

    def got_error(failure):
        log_failure(failure)

    connector = RemoteBrokerConnector(reactor, config)
    result = connector.connect()
    result.addCallback(got_connection)
    result.addErrback(got_error)
    result.addBoth(lambda x: connector.disconnect())

    # For some obscure reason our LandscapeReactor.stop method calls
    # reactor.crash() instead of reactor.stop(), which doesn't work
    # here. Maybe LandscapeReactor.stop should simply use reactor.stop().
    result.addBoth(lambda ignored: reactor.call_later(
        0, reactor._reactor.stop))

    reactor.run()

    return result



import os
import time

from landscape.tests.helpers import LandscapeTest

from landscape.lib.fs import append_file, read_file, touch_file


class ReadFileTest(LandscapeTest):

    def test_read_file(self):
        """
        With no options L{read_file} reads the whole file passed as argument.
        """
        path = self.makeFile("foo")
        self.assertEqual(read_file(path), "foo")

    def test_read_file_with_limit(self):
        """
        With a positive limit L{read_file} reads only the bytes after the
        given limit.
        """
        path = self.makeFile("foo bar")
        self.assertEqual(read_file(path, limit=3), " bar")

    def test_read_file_with_negative_limit(self):
        """
        With a negative limit L{read_file} reads only the tail of the file.
        """
        path = self.makeFile("foo bar from end")
        self.assertEqual(read_file(path, limit=-3), "end")

    def test_read_file_with_limit_bigger_than_file(self):
        """
        If the limit is bigger than the file L{read_file} reads the entire
        file.
        """
        path = self.makeFile("foo bar from end")
        self.assertEqual(read_file(path, limit=100), "foo bar from end")
        self.assertEqual(read_file(path, limit=-100), "foo bar from end")


class TouchFileTest(LandscapeTest):

    def test_touch_file(self):
        """
        The L{touch_file} function touches a file, setting its last
        modification time.
        """
        path = self.makeFile()
        utime_mock = self.mocker.replace("os.utime")
        self.expect(utime_mock(path, None))
        self.mocker.replay()
        touch_file(path)
        self.assertFileContent(path, "")

    def test_touch_file_multiple_times(self):
        """
        The L{touch_file} function can be called multiple times.
        """
        path = self.makeFile()
        touch_file(path)
        touch_file(path)
        self.assertFileContent(path, "")

    def test_touch_file_with_offset_seconds(self):
        """
        The L{touch_file} function can be called with a offset in seconds that
        will be reflected in the access and modification times of the file.
        """
        path = self.makeFile()
        current_time = long(time.time())
        expected_time = current_time - 1
        time_mock = self.mocker.replace("time.time")
        self.expect(time_mock())
        self.mocker.result(current_time)
        utime_mock = self.mocker.replace("os.utime")
        self.expect(utime_mock(path, (expected_time, expected_time)))
        self.mocker.replay()
        touch_file(path, offset_seconds=-1)
        self.assertFileContent(path, "")


class AppendFileTest(LandscapeTest):

    def test_append_existing_file(self):
        """
        The L{append_file} function appends contents to an existing file.
        """
        existing_file = self.makeFile("foo bar")
        append_file(existing_file, " baz")
        self.assertFileContent(existing_file, "foo bar baz")

    def test_append_no_file(self):
        """
        The L{append_file} function creates a new file if one doesn't
        exist already.
        """
        new_file = os.path.join(self.makeDir(), "new_file")
        append_file(new_file, "contents")
        self.assertFileContent(new_file, "contents")



import unittest

from landscape.lib.sequenceranges import (
    SequenceRanges, remove_from_ranges, add_to_ranges, find_ranges_index,
    ranges_to_sequence, sequence_to_ranges, SequenceError)


class SequenceRangesTest(unittest.TestCase):

    def setUp(self):
        self.ranges = [1, 2, (15, 17), 19, (21, 24), 26, 27]
        self.sequence = [1, 2, 15, 16, 17, 19, 21, 22, 23, 24, 26, 27]

    def test_empty_to_sequence(self):
        self.assertEqual(SequenceRanges().to_sequence(), [])

    def test_empty_to_ranges(self):
        self.assertEqual(SequenceRanges().to_ranges(), [])

    def test_from_to_sequence(self):
        obj = SequenceRanges.from_sequence(self.sequence)
        self.assertEqual(obj.to_sequence(), self.sequence)

    def test_from_to_ranges(self):
        obj = SequenceRanges.from_ranges(self.ranges)
        self.assertEqual(obj.to_ranges(), self.ranges)

    def test_to_ranges_immutable(self):
        obj = SequenceRanges.from_ranges(self.ranges)
        obj.to_ranges().append(123)
        self.assertEqual(obj.to_ranges(), self.ranges)

    def test_from_sequence_to_ranges(self):
        obj = SequenceRanges.from_sequence(self.sequence)
        self.assertEqual(list(obj.to_ranges()), self.ranges)

    def test_from_ranges_to_sequence(self):
        obj = SequenceRanges.from_ranges(self.ranges)
        self.assertEqual(list(obj.to_sequence()), self.sequence)

    def test_iter(self):
        obj = SequenceRanges.from_ranges(self.ranges)
        self.assertEqual(list(obj), self.sequence)

    def test_contains(self):
        obj = SequenceRanges.from_ranges(self.ranges)
        self.assertTrue(1 in obj)
        self.assertTrue(2 in obj)
        self.assertTrue(15 in obj)
        self.assertTrue(16 in obj)
        self.assertTrue(17 in obj)
        self.assertTrue(19 in obj)
        self.assertTrue(27 in obj)
        self.assertTrue(0 not in obj)
        self.assertTrue(3 not in obj)
        self.assertTrue(14 not in obj)
        self.assertTrue(18 not in obj)
        self.assertTrue(20 not in obj)
        self.assertTrue(28 not in obj)

    def test_add(self):
        obj = SequenceRanges()
        obj.add(1)
        self.assertEqual(obj.to_ranges(), [1])
        obj.add(2)
        self.assertEqual(obj.to_ranges(), [1, 2])
        obj.add(3)
        self.assertEqual(obj.to_ranges(), [(1, 3)])
        obj.add(3)
        self.assertEqual(obj.to_ranges(), [(1, 3)])

    def test_remove(self):
        obj = SequenceRanges.from_ranges([(1, 3)])
        obj.remove(2)
        self.assertEqual(obj.to_ranges(), [1, 3])
        obj.remove(1)
        self.assertEqual(obj.to_ranges(), [3])
        obj.remove(3)
        self.assertEqual(obj.to_ranges(), [])
        obj.remove(4)
        self.assertEqual(obj.to_ranges(), [])


class SequenceToRangesTest(unittest.TestCase):

    def test_empty(self):
        self.assertEqual(list(sequence_to_ranges([])), [])

    def test_one_element(self):
        self.assertEqual(list(sequence_to_ranges([1])), [1])

    def test_two_elements(self):
        self.assertEqual(list(sequence_to_ranges([1, 2])), [1, 2])

    def test_three_elements(self):
        self.assertEqual(list(sequence_to_ranges([1, 2, 3])), [(1, 3)])

    def test_many_elements(self):
        sequence = [1, 2, 15, 16, 17, 19, 21, 22, 23, 24, 26, 27]
        self.assertEqual(list(sequence_to_ranges(sequence)),
                         [1, 2, (15, 17), 19, (21, 24), 26, 27])

    def test_out_of_order(self):
        self.assertRaises(SequenceError, sequence_to_ranges([2, 1]).next)

    def test_duplicated_item(self):
        self.assertRaises(SequenceError, sequence_to_ranges([1, 1]).next)


class RangesToSequenceTest(unittest.TestCase):

    def test_empty(self):
        self.assertEqual(list(ranges_to_sequence([])), [])

    def test_one_element(self):
        self.assertEqual(list(ranges_to_sequence([1])), [1])

    def test_two_elements(self):
        self.assertEqual(list(ranges_to_sequence([1, 2])), [1, 2])

    def test_three_elements(self):
        self.assertEqual(list(ranges_to_sequence([(1, 3)])), [1, 2, 3])

    def test_many_elements(self):
        ranges = [1, 2, (15, 17), 19, (21, 24), 26, 27]
        self.assertEqual(list(ranges_to_sequence(ranges)),
                         [1, 2, 15, 16, 17, 19, 21, 22, 23, 24, 26, 27])

    def test_invalid_range(self):
        """
        If range start value is greater than the end one, an error is raised.
        """
        ranges = [1, 2, (5, 3), 10]
        self.assertRaises(ValueError, list, ranges_to_sequence(ranges))


class FindRangesIndexTest(unittest.TestCase):

    def test_empty(self):
        self.assertEqual(find_ranges_index([], 2), 0)

    def test_sequence(self):
        self.assertEqual(find_ranges_index([1, 2, 3, 4, 5], 0), 0)
        self.assertEqual(find_ranges_index([1, 2, 3, 4, 5], 1), 0)
        self.assertEqual(find_ranges_index([1, 2, 3, 4, 5], 2), 1)
        self.assertEqual(find_ranges_index([1, 2, 3, 4, 5], 3), 2)
        self.assertEqual(find_ranges_index([1, 2, 3, 4, 5], 4), 3)
        self.assertEqual(find_ranges_index([1, 2, 3, 4, 5], 5), 4)
        self.assertEqual(find_ranges_index([1, 2, 3, 4, 5], 6), 5)

    def test_sequence_with_missing(self):
        self.assertEqual(find_ranges_index([1, 2, 4, 5], 2), 1)
        self.assertEqual(find_ranges_index([1, 2, 4, 5], 3), 2)
        self.assertEqual(find_ranges_index([1, 2, 4, 5], 4), 2)

    def test_range(self):
        self.assertEqual(find_ranges_index([1, (2, 4), 5], 0), 0)
        self.assertEqual(find_ranges_index([1, (2, 4), 5], 1), 0)
        self.assertEqual(find_ranges_index([1, (2, 4), 5], 2), 1)
        self.assertEqual(find_ranges_index([1, (2, 4), 5], 3), 1)
        self.assertEqual(find_ranges_index([1, (2, 4), 5], 4), 1)
        self.assertEqual(find_ranges_index([1, (2, 4), 5], 5), 2)
        self.assertEqual(find_ranges_index([1, (2, 4), 5], 6), 3)

    def test_range_with_missing(self):
        self.assertEqual(find_ranges_index([1, (3, 4), 5], 0), 0)
        self.assertEqual(find_ranges_index([1, (3, 4), 5], 1), 0)
        self.assertEqual(find_ranges_index([1, (3, 4), 5], 2), 1)
        self.assertEqual(find_ranges_index([1, (3, 4), 5], 3), 1)
        self.assertEqual(find_ranges_index([1, (3, 4), 5], 4), 1)
        self.assertEqual(find_ranges_index([1, (3, 4), 5], 5), 2)
        self.assertEqual(find_ranges_index([1, (3, 4), 5], 6), 3)


class AddToRangesTest(unittest.TestCase):

    def test_empty(self):
        ranges = []
        add_to_ranges(ranges, 1)
        self.assertEqual(ranges, [1])

    def test_append(self):
        ranges = [1]
        add_to_ranges(ranges, 2)
        self.assertEqual(ranges, [1, 2])

    def test_prepend(self):
        ranges = [2]
        add_to_ranges(ranges, 1)
        self.assertEqual(ranges, [1, 2])

    def test_insert(self):
        ranges = [1, 4]
        add_to_ranges(ranges, 2)
        self.assertEqual(ranges, [1, 2, 4])

    def test_merge_sequence(self):
        ranges = [1, 2, 4, 5]
        add_to_ranges(ranges, 3)
        self.assertEqual(ranges, [(1, 5)])

    def test_merge_ranges(self):
        ranges = [(1, 3), (5, 7)]
        add_to_ranges(ranges, 4)
        self.assertEqual(ranges, [(1, 7)])

    def test_merge_sequence_and_ranges(self):
        ranges = [(1, 3), 5, 6, 7]
        add_to_ranges(ranges, 4)
        self.assertEqual(ranges, [(1, 7)])

    def test_merge_sequence_and_ranges_with_gaps(self):
        ranges = [1, (3, 5), 7, 9]
        add_to_ranges(ranges, 6)
        self.assertEqual(ranges, [1, (3, 7), 9])

    def test_dont_merge_ranges_with_gap(self):
        ranges = [(1, 3), (7, 9)]
        add_to_ranges(ranges, 5)
        self.assertEqual(ranges, [(1, 3), 5, (7, 9)])

    def test_duplicate(self):
        ranges = [1]
        add_to_ranges(ranges, 1)
        self.assertEqual(ranges, [1])

    def test_duplicate_in_range(self):
        ranges = [(1, 3)]
        add_to_ranges(ranges, 1)
        self.assertEqual(ranges, [(1, 3)])
        add_to_ranges(ranges, 2)
        self.assertEqual(ranges, [(1, 3)])
        add_to_ranges(ranges, 3)
        self.assertEqual(ranges, [(1, 3)])


class RemoveFromRangesTest(unittest.TestCase):

    def test_empty(self):
        ranges = []
        remove_from_ranges(ranges, 1)
        self.assertEqual(ranges, [])

    def test_single(self):
        ranges = [1]
        remove_from_ranges(ranges, 1)
        self.assertEqual(ranges, [])

    def test_remove_before(self):
        ranges = [1, 2]
        remove_from_ranges(ranges, 1)
        self.assertEqual(ranges, [2])

    def test_remove_after(self):
        ranges = [1, 2]
        remove_from_ranges(ranges, 2)
        self.assertEqual(ranges, [1])

    def test_remove_inside(self):
        ranges = [1, 2, 3]
        remove_from_ranges(ranges, 2)
        self.assertEqual(ranges, [1, 3])

    def test_remove_unexistent(self):
        ranges = [1, 3]
        remove_from_ranges(ranges, 2)
        self.assertEqual(ranges, [1, 3])

    def test_split_range(self):
        ranges = [(1, 5)]
        remove_from_ranges(ranges, 3)
        self.assertEqual(ranges, [1, 2, 4, 5])

    def test_split_range_into_ranges(self):
        ranges = [(1, 7)]
        remove_from_ranges(ranges, 4)
        self.assertEqual(ranges, [(1, 3), (5, 7)])

    def test_decrement_left(self):
        ranges = [(1, 5)]
        remove_from_ranges(ranges, 1)
        self.assertEqual(ranges, [(2, 5)])

    def test_decrement_right(self):
        ranges = [(1, 5)]
        remove_from_ranges(ranges, 5)
        self.assertEqual(ranges, [(1, 4)])

    def test_dont_removing_unmatched_range(self):
        ranges = [(1, 3), (5, 7)]
        remove_from_ranges(ranges, 4)
        self.assertEqual(ranges, [(1, 3), (5, 7)])


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(SequenceToRangesTest),
        unittest.makeSuite(RangesToSequenceTest),
        unittest.makeSuite(SequenceRangesTest),
        unittest.makeSuite(FindRangesIndexTest),
        unittest.makeSuite(AddToRangesTest),
        unittest.makeSuite(RemoveFromRangesTest),
    ))



"""The Landscape monitor plugin system."""

import os

from landscape.broker.client import BrokerClient


class Monitor(BrokerClient):
    """The central point of integration in the Landscape monitor."""

    name = "monitor"

    def __init__(self, reactor, config, persist, persist_filename=None,
                 step_size=5 * 60):
        super(Monitor, self).__init__(reactor)
        self.reactor = reactor
        self.config = config
        self.persist = persist
        self.persist_filename = persist_filename
        if persist_filename and os.path.exists(persist_filename):
            self.persist.load(persist_filename)
        self._plugins = []
        self.step_size = step_size
        self.reactor.call_every(self.config.flush_interval, self.flush)

    def flush(self):
        """Flush data to disk."""
        if self.persist_filename:
            self.persist.save(self.persist_filename)

    def exchange(self):
        """Call C{exchange} on all plugins."""
        super(Monitor, self).exchange()
        self.flush()



import dbus

import landscape.ui.model.registration.mechanism as mechanism


class RegistrationProxy(object):
    """
    L{RegistrationProxy} allows the use of the L{RegistrationMechanism} via
    DBus without having to know about DBus.  This in turn allows controller
    code to remain agnostic to the implementation of registration.
    """

    def __init__(self, on_register_notify=None, on_register_error=None,
                 on_register_succeed=None, on_register_fail=None,
                 on_disable_succeed=None, on_disable_fail=None, bus=None):
        self._bus = None
        self._interface = None
        self._on_register_notify = on_register_notify
        self._on_register_error = on_register_error
        self._on_register_succeed = on_register_succeed
        self._on_register_fail = on_register_fail
        self._on_disable_succeed = on_disable_succeed
        self._on_disable_fail = on_disable_fail
        self._setup_interface(bus)

    def _setup_interface(self, bus):
        """
        Redefining L{_setup_interface} allows us to bypass DBus for more
        convenient testing in some instances.
        """
        if bus is None:
            self._bus = dbus.SystemBus()
        else:
            self._bus = bus
        self._remote_object = self._bus.get_object(mechanism.SERVICE_NAME,
                                                   mechanism.OBJECT_PATH)
        self._interface = dbus.Interface(self._remote_object,
                                         mechanism.INTERFACE_NAME)

    def _exit_handler_wrapper(self, exit_handler):

        def wrapped_exit_handler(message):
            self._remove_handlers()
            exit_handler(message)

        return wrapped_exit_handler

    def _register_handlers(self):
        self._handlers = []
        if self._on_register_notify:
            self._handlers.append(
                self._bus.add_signal_receiver(
                    self._on_register_notify,
                    signal_name="register_notify",
                    dbus_interface=mechanism.INTERFACE_NAME,
                    bus_name=None,
                    path=mechanism.OBJECT_PATH))
        if self._on_register_error:
            self._handlers.append(
                self._bus.add_signal_receiver(
                    self._on_register_error,
                    signal_name="register_error",
                    dbus_interface=mechanism.INTERFACE_NAME,
                    bus_name=None,
                    path=mechanism.OBJECT_PATH))
        if self._on_register_succeed:
            self._handlers.append(
                self._bus.add_signal_receiver(
                    self._exit_handler_wrapper(self._on_register_succeed),
                    signal_name="register_succeed",
                    dbus_interface=mechanism.INTERFACE_NAME,
                    bus_name=None,
                    path=mechanism.OBJECT_PATH))
        if self._on_register_fail:
            self._handlers.append(
                self._bus.add_signal_receiver(
                    self._exit_handler_wrapper(self._on_register_fail),
                    signal_name="register_fail",
                    dbus_interface=mechanism.INTERFACE_NAME,
                    bus_name=None,
                    path=mechanism.OBJECT_PATH))
        if self._on_disable_succeed:
            self._handlers.append(
                self._bus.add_signal_receiver(
                    self._exit_handler_wrapper(self._on_disable_succeed),
                    signal_name="disable_succeed",
                    dbus_interface=mechanism.INTERFACE_NAME,
                    bus_name=None,
                    path=mechanism.OBJECT_PATH))
        if self._on_disable_fail:
            self._handlers.append(
                self._bus.add_signal_receiver(
                    self._exit_handler_wrapper(self._on_disable_fail),
                    signal_name="disable_fail",
                    dbus_interface=mechanism.INTERFACE_NAME,
                    bus_name=None,
                    path=mechanism.OBJECT_PATH))

    def _remove_handlers(self):
        for handler in self._handlers:
            self._bus.remove_signal_receiver(handler)

    def challenge(self):
        return self._interface.challenge()

    def register(self, config_path):
        self._register_handlers()
        try:
            result, message = self._interface.register(config_path)
        except dbus.DBusException, e:
            if e.get_dbus_name() != "org.freedesktop.DBus.Error.NoReply":
                raise
            else:
                result = False
                message = "Registration timed out."
        if result:
            self._on_register_succeed()
        else:
            self._on_register_error(message)
        return result

    def disable(self):
        self._register_handlers()
        result = self._interface.disable()
        if result:
            self._on_disable_succeed()
        else:
            self._on_disable_fail()
        return result

    def exit(self):
        """
        Cause the mechanism to exit.
        """
        try:
            self._interface.exit()
        except dbus.DBusException, e:
            if e.get_dbus_name() != "org.freedesktop.DBus.Error.NoReply":
                raise



from collections import namedtuple
import json

from landscape.tests.helpers import LandscapeTest
from landscape.lib.juju import get_juju_info


SAMPLE_JUJU_INFO = json.dumps({"environment-uuid": "DEAD-BEEF",
                               "unit-name": "service/0",
                               "api-addresses": "10.0.3.1:17070",
                               "private-address": "127.0.0.1"})


class JujuTest(LandscapeTest):

    Config = namedtuple("Config", "juju_filename")

    def test_get_juju_info_sample_data(self):
        """L{get_juju_info} parses JSON data from the juju_filename file."""
        stub_config = self.Config(self.makeFile(SAMPLE_JUJU_INFO))
        juju_info = get_juju_info(stub_config)
        self.assertEqual(
            {u"environment-uuid": "DEAD-BEEF",
             u"unit-name": "service/0",
             u"api-addresses": ["10.0.3.1:17070"],
             u"private-address": "127.0.0.1"}, juju_info)

    def test_get_juju_info_empty_file(self):
        """
        If L{get_juju_info} is called with a configuration pointing to
        an empty file, it returns C{None}.
        """
        stub_config = self.Config(self.makeFile(""))
        juju_info = get_juju_info(stub_config)
        self.log_helper.ignore_errors(ValueError)
        self.assertEqual(juju_info, None)
        self.assertIn("Error attempting to read JSON", self.logfile.getvalue())

    def test_get_juju_info_not_a_file(self):
        """
        If L{get_juju_info} is called with a configuration pointing to
        a directory, it returns C{None}.
        """
        stub_config = self.Config("/")
        juju_info = get_juju_info(stub_config)
        self.assertIs(juju_info, None)

    def test_get_juju_info_multiple_endpoints(self):
        """L{get_juju_info} turns space separated API addresses into a list."""
        juju_multiple_endpoints = json.dumps({
                "environment-uuid": "DEAD-BEEF",
                "unit-name": "service/0",
                "api-addresses": "10.0.3.1:17070 10.0.3.2:18080",
                "private-address": "127.0.0.1"})

        stub_config = self.Config(self.makeFile(juju_multiple_endpoints))
        juju_info = get_juju_info(stub_config)
        self.assertEqual(
            {u"environment-uuid": "DEAD-BEEF",
             u"unit-name": "service/0",
             u"api-addresses": ["10.0.3.1:17070", "10.0.3.2:18080"],
             u"private-address": "127.0.0.1"}, juju_info)



from twisted.internet.defer import succeed

from landscape.lib.fetch import HTTPCodeError
from landscape.monitor.swiftdeviceinfo import SwiftDeviceInfo
from landscape.tests.helpers import LandscapeTest, MonitorHelper
from landscape.tests.mocker import ANY


class FakeRingInfo(object):
    def __init__(self, ip_port_tuples=[]):
        self.devs = []
        for ip, port in ip_port_tuples:
            self.devs.append({"ip": ip, "port": port})


class SwiftDeviceInfoTest(LandscapeTest):
    """Tests for swift-device-info plugin."""

    helpers = [MonitorHelper]

    def setUp(self):
        LandscapeTest.setUp(self)
        self.mstore.set_accepted_types(["swift-device-info"])

    def test_exchange_messages(self):
        """
        The swift_device_info plugin queues message when manager.exchange()
        is called.  Each message should be aligned to a step boundary;
        only a sing message with the latest swift device information will
        be delivered in a single message.
        """
        def fake_swift_devices():
            return [{"device": "/dev/hdf", "mounted": True},
                    {"device": "/dev/hda2", "mounted": False}]

        plugin = SwiftDeviceInfo(create_time=self.reactor.time)
        plugin._get_swift_devices = fake_swift_devices

        step_size = self.monitor.step_size
        self.monitor.add(plugin)

        # Exchange should trigger a flush of the persist database
        registry_mocker = self.mocker.replace(plugin.registry)
        registry_mocker.flush()
        self.mocker.result(None)
        self.mocker.replay()

        self.reactor.advance(step_size * 2)
        self.monitor.exchange()

        messages = self.mstore.get_pending_messages()
        self.assertEqual(len(messages), 1)
        expected_message_content = [
                {"device": "/dev/hdf", "mounted": True},
                {"device": "/dev/hda2", "mounted": False}]

        swift_devices = messages[0]["swift-device-info"]
        self.assertEqual(swift_devices, expected_message_content)

    def test_messaging_flushes(self):
        """
        Duplicate message should never be created.  If no data is
        available, None will be returned when messages are created.
        """
        def fake_swift_devices():
            return [{"device": "/dev/hdf", "mounted": True},
                    {"device": "/dev/hda2", "mounted": False}]

        plugin = SwiftDeviceInfo(create_time=self.reactor.time)
        self.monitor.add(plugin)
        plugin._get_swift_devices = fake_swift_devices

        self.reactor.advance(self.monitor.step_size)

        message = plugin.create_swift_device_info_message()
        self.assertEqual(message.keys(), ["swift-device-info", "type"])

        message = plugin.create_swift_device_info_message()
        self.assertEqual(message, None)

    def test_never_exchange_empty_messages(self):
        """
        When the plugin has no data, its various create_X_message()
        methods will return None.  Empty or null messages should never
        be queued.
        """
        self.mstore.set_accepted_types(["load-average"])

        plugin = SwiftDeviceInfo()
        self.monitor.add(plugin)
        self.monitor.exchange()
        self.assertEqual(len(self.mstore.get_pending_messages()), 0)

    def test_messages_with_swift_data(self):
        """
        All swift-affiliated devices are sent in swift-device-info messages.
        Both mounted and unmounted swift devices send data.
        """
        def fake_swift_devices():
            return [{"device": "/dev/hdf", "mounted": True},
                    {"device": "/dev/hda2", "mounted": False}]

        plugin = SwiftDeviceInfo(create_time=self.reactor.time)

        plugin._get_swift_devices = fake_swift_devices

        step_size = self.monitor.step_size
        self.monitor.add(plugin)
        plugin.run()

        self.reactor.advance(step_size)
        self.monitor.exchange()

        messages = self.mstore.get_pending_messages()
        self.assertEqual(len(messages), 1)

        # Need to see both mounted and unmounted swift device info
        self.assertEqual(
            messages[0].get("swift-device-info"),
            [{'device': u'/dev/hdf', 'mounted': True},
             {'device': u'/dev/hda2', 'mounted': False}])

    def test_resynchronize(self):
        """
        On the reactor "resynchronize" event, new swift-device-info messages
        should be sent.
        """
        plugin = SwiftDeviceInfo(create_time=self.reactor.time)

        def fake_swift_devices():
            return [{"device": "/dev/hdf", "mounted": True},
                    {"device": "/dev/hda2", "mounted": False}]

        self.monitor.add(plugin)
        plugin._get_swift_devices = fake_swift_devices

        plugin.run()
        plugin.exchange()
        self.reactor.fire("resynchronize", scopes=["storage"])
        plugin.run()
        plugin.exchange()
        messages = self.mstore.get_pending_messages()
        expected_message = {
            "type": "swift-device-info",
            "swift-device-info": [
                {"device": "/dev/hdf", "mounted": True},
                {"device": "/dev/hda2", "mounted": False}]}
        self.assertMessages(messages, [expected_message, expected_message])

    def test_no_message_if_not_accepted(self):
        """
        Don't add any messages at all if the broker isn't currently
        accepting their type.
        """
        self.mstore.set_accepted_types([])

        def fake_swift_devices():
            return [{"device": "/dev/hdf", "mounted": True},
                    {"device": "/dev/hda2", "mounted": False}]

        plugin = SwiftDeviceInfo(create_time=self.reactor.time)
        self.monitor.add(plugin)
        plugin._get_swift_devices = fake_swift_devices

        self.reactor.advance(self.monitor.step_size * 2)
        self.monitor.exchange()

        self.mstore.set_accepted_types(["swift-device-info"])
        self.assertMessages(list(self.mstore.get_pending_messages()), [])

    def test_call_on_accepted(self):
        """
        When message type acceptance is added for swift-device-info,
        send_message gets called.
        """
        def fake_swift_devices():
            return [{"device": "/dev/hdf", "mounted": True},
                    {"device": "/dev/hda2", "mounted": False}]

        plugin = SwiftDeviceInfo(create_time=self.reactor.time)
        self.monitor.add(plugin)
        plugin._get_swift_devices = fake_swift_devices

        self.reactor.advance(plugin.run_interval)

        remote_broker_mock = self.mocker.replace(self.remote)
        remote_broker_mock.send_message(ANY, ANY, urgent=True)
        self.mocker.result(succeed(None))
        self.mocker.count(1)  # 1 send message is called for swift-device-info
        self.mocker.replay()

        self.reactor.fire(
            ("message-type-acceptance-changed", "swift-device-info"), True)

    def test_persist_deltas(self):
        """
        Swift persistent device info drops old devices from persist storage if
        the device no longer exists in the current device list.
        """
        def fake_swift_devices():
            return [{"device": "/dev/hdf", "mounted": True},
                    {"device": "/dev/hda2", "mounted": False}]

        def fake_swift_devices_no_hdf():
            return [{"device": "/dev/hda2", "mounted": False}]

        plugin = SwiftDeviceInfo(create_time=self.reactor.time)
        self.monitor.add(plugin)
        plugin._get_swift_devices = fake_swift_devices
        plugin.run()
        plugin.exchange()  # To persist swift recon data
        self.assertEqual(
            plugin._persist.get("swift-device-info"),
            {"/dev/hdf": {"device": "/dev/hdf", "mounted": True},
             "/dev/hda2": {"device": "/dev/hda2", "mounted": False}})

        # Drop a device
        plugin._get_swift_devices = fake_swift_devices_no_hdf
        plugin.run()
        plugin.exchange()
        self.assertEqual(
            plugin._persist.get("swift-device-info"),
            {"/dev/hda2": {"device": "/dev/hda2", "mounted": False}})

        # Run again, calling create_swift_device_info_message which purges info
        plugin.run()
        plugin.exchange()
        message3 = plugin.create_swift_device_info_message()
        self.assertIdentical(message3, None)

    def test_persist_timing(self):
        """Swift device info is only persisted when exchange happens.

        If an event happened between the persist and the exchange, the server
        didn't get the mount info at all. This test ensures that mount info are
        only saved when exchange happens.
        """
        def fake_swift_devices():
            return [{"device": "/dev/hdf", "mounted": True},
                    {"device": "/dev/hda2", "mounted": False}]

        plugin = SwiftDeviceInfo(create_time=self.reactor.time)
        self.monitor.add(plugin)
        plugin._get_swift_devices = fake_swift_devices
        plugin.run()
        message1 = plugin.create_swift_device_info_message()
        self.assertEqual(
            message1.get("swift-device-info"),
            [{"device": "/dev/hdf", "mounted": True},
             {"device": "/dev/hda2", "mounted": False}])
        plugin.run()
        message2 = plugin.create_swift_device_info_message()
        self.assertEqual(
            message2.get("swift-device-info"),
            [{"device": "/dev/hdf", "mounted": True},
             {"device": "/dev/hda2", "mounted": False}])
        # Run again, calling create_swift_device_info_message which purges info
        plugin.run()
        plugin.exchange()
        plugin.run()
        message3 = plugin.create_swift_device_info_message()
        self.assertIdentical(message3, None)

    def test_wb_get_swift_devices_when_not_a_swift_node(self):
        """
        When not a swift node, _get_swift_devices returns an empty list and
        no error messages.
        """
        plugin = SwiftDeviceInfo(create_time=self.reactor.time)
        self.assertEqual(plugin._get_swift_devices(), [])

    def test_wb_get_swift_devices_when_on_a_swift_node(self):
        """
        When on a swift node, _get_swift_devices reports a warning if the ring
        files don't exist yet.
        """
        plugin = SwiftDeviceInfo(create_time=self.reactor.time,
                                     swift_config="/etc/hosts")
        logging_mock = self.mocker.replace("logging.warning")
        logging_mock("Swift ring files are not available yet.")
        self.mocker.replay()
        self.assertEqual(plugin._get_swift_devices(), [])

    def test_run_disabled_when_missing_swift_config(self):
        """
        When on a node that doesn't have the appropriate swift config file. The
        plugin logs an info message and is disabled.
        """
        plugin = SwiftDeviceInfo(create_time=self.reactor.time,
                                     swift_config="/config/file/doesnotexist")
        logging_mock = self.mocker.replace("logging.info")
        logging_mock("This does not appear to be a swift storage server. "
                     "'swift-device-info' plugin has been disabled.")
        self.mocker.replay()
        self.monitor.add(plugin)
        self.assertEqual(plugin.enabled, True)
        plugin.run()
        self.assertEqual(plugin.enabled, False)

    def test_wb_get_swift_devices_no_swift_python_libs_available(self):
        """
        The plugin logs an error and doesn't find swift devices when it can't
        import the swift python libs which it requires.
        """
        plugin = SwiftDeviceInfo(create_time=self.reactor.time,
                                     swift_config="/etc/hosts",
                                     swift_ring="/etc/hosts")

        logging_mock = self.mocker.replace("logging.error")
        logging_mock("Swift python common libraries not found. "
                     "'swift-device-info' plugin has been disabled.")
        self.mocker.replay()

        self.assertEqual(plugin._get_swift_devices(), [])

    def test_wb_get_swift_disk_usage_when_no_swift_service_running(self):
        """
        When the swift service is running, but recon middleware is not active,
        the Swift storage usage logs an error.
        """
        self.log_helper.ignore_errors(".*")
        plugin = SwiftDeviceInfo(create_time=self.reactor.time)
        plugin._swift_recon_url = "http://localhost:12654"
        result = plugin._get_swift_disk_usage()
        self.assertIs(None, result)
        self.assertIn(
            "Swift service not available at %s." % plugin._swift_recon_url,
            self.logfile.getvalue())

    def test_wb_get_swift_disk_usage_when_no_recon_service_configured(self):
        """
        When the swift service is running, but recon middleware is not active,
        an error is logged.
        """
        plugin = SwiftDeviceInfo(create_time=self.reactor.time)

        plugin._swift_recon_url = "http://localhost:12654"

        def fetch_error(url):
            raise HTTPCodeError(400, "invalid path: /recon/diskusage")
        plugin._fetch = fetch_error

        logging_mock = self.mocker.replace("logging.error", passthrough=False)
        logging_mock(
            "Swift service is running without swift-recon enabled. "
            "'swift-device-info' plugin has been disabled.")
        self.mocker.result(None)
        self.mocker.replay()

        result = plugin._get_swift_disk_usage()
        self.assertIs(None, result)

    def test_wb_get_swift_usage_no_information(self):
        """
        When the swift recon service returns no disk usage information,
        the _get_swift_disk_usage method returns None.
        """
        plugin = SwiftDeviceInfo(create_time=self.reactor.time)

        def fetch_none(url):
            return None

        plugin._fetch = fetch_none

        result = plugin._get_swift_disk_usage()
        self.assertEqual(None, result)

    def test_wb_get_swift_devices_no_matched_local_service(self):
        """
        The plugin logs an error when the swift ring file does not represent
        a swift service running local IP address on the current node.
        """
        plugin = SwiftDeviceInfo(create_time=self.reactor.time,
                                     swift_config="/etc/hosts")

        def get_fake_ring():
            return FakeRingInfo([("192.168.1.10", 6000)])
        plugin._get_ring = get_fake_ring

        def local_network_devices():
            return [{"ip_address": "10.1.2.3"}]
        plugin._get_network_devices = local_network_devices

        logging_mock = self.mocker.replace("logging.error")
        logging_mock("Local swift service not found. "
                     "'swift-device-info' plugin has been disabled.")
        self.mocker.replay()
        self.assertEqual(plugin._get_swift_devices(), [])



from landscape.tests.helpers import (
    LandscapeTest, FakeBrokerServiceHelper)
from landscape.reactor import FakeReactor
from landscape.manager.config import ManagerConfiguration, ALL_PLUGINS
from landscape.manager.service import ManagerService
from landscape.manager.processkiller import ProcessKiller


class ManagerServiceTest(LandscapeTest):

    helpers = [FakeBrokerServiceHelper]

    def setUp(self):
        super(ManagerServiceTest, self).setUp()
        config = ManagerConfiguration()
        config.load(["-c", self.config_filename])

        class FakeManagerService(ManagerService):
            reactor_factory = FakeReactor

        self.service = FakeManagerService(config)

    def test_plugins(self):
        """
        By default the L{ManagerService.plugins} list holds an instance of
        every enabled manager plugin.
        """
        self.assertEqual(len(self.service.plugins), len(ALL_PLUGINS))

    def test_get_plugins(self):
        """
        If the C{--manager-plugins} command line option is specified, only the
        given plugins will be enabled.
        """
        self.service.config.load(["--manager-plugins", "ProcessKiller"])
        [plugin] = self.service.get_plugins()
        self.assertTrue(isinstance(plugin, ProcessKiller))

    def test_start_service(self):
        """
        The L{ManagerService.startService} method connects to the broker,
        starts the plugins and register the manager as broker client.
        """
        def stop_service(ignored):
            for plugin in self.service.plugins:
                if getattr(plugin, "stop", None) is not None:
                    plugin.stop()
            [connector] = self.broker_service.broker.get_connectors()
            connector.disconnect()
            self.service.stopService()
            self.broker_service.stopService()

        def assert_broker_connection(ignored):
            self.assertEqual(len(self.broker_service.broker.get_clients()), 1)
            self.assertIs(self.service.broker, self.service.manager.broker)
            result = self.service.broker.ping()
            return result.addCallback(stop_service)

        self.broker_service.startService()
        started = self.service.startService()
        return started.addCallback(assert_broker_connection)



from datetime import datetime

from landscape.lib.timestamp import to_timestamp
from landscape.tests.helpers import LandscapeTest


class TimestampTest(LandscapeTest):
    """Test for timestamp conversion function."""

    def test_conversion(self):
        """Test ensures that the conversion returns an int, not a float."""
        date = datetime.utcfromtimestamp(1000)
        timestamp = to_timestamp(date)
        self.assertTrue(isinstance(timestamp, int))
        self.assertEqual(timestamp, 1000)

    def test_before_epoch_conversion(self):
        """Test converting a date before the epoch."""
        date = datetime.utcfromtimestamp(-1000)
        self.assertEqual(to_timestamp(date), -1000)



import logging

from landscape.lib.juju import get_juju_info
from landscape.monitor.plugin import MonitorPlugin


class JujuInfo(MonitorPlugin):
    """Plugin for reporting Juju information."""

    persist_name = "juju-info"
    scope = "juju"

    def register(self, registry):
        super(JujuInfo, self).register(registry)
        self.call_on_accepted("juju-info", self.send_juju_message, True)

    def exchange(self, urgent=False):
        broker = self.registry.broker
        broker.call_if_accepted("juju-info", self.send_juju_message, urgent)

    def send_juju_message(self, urgent=False):
        message = self._create_juju_info_message()
        if message:
            message["type"] = "juju-info"
            logging.info("Queuing message with updated juju info.")
            self.registry.broker.send_message(message, self._session_id,
                                              urgent=urgent)

    def _create_juju_info_message(self):
        message = get_juju_info(self.registry.config)
        if message != self._persist.get("juju-info"):
            self._persist.set("juju-info", message)
            return message
        return None



"""Expose the methods of a remote object over AMP.

This module implements an AMP-based protocol for performing remote procedure
calls in a convenient and easy way. It's conceptually similar to DBus in that
it supports exposing a Python object to a remote process, with communication
happening over any Twisted-supported transport, e.g. Unix domain sockets.

For example let's say we have a Python process "A" that creates an instance of
this class::

    class Greeter(object):

        def hello(self, name):
            return "hi %s!" % name

    greeter = Greeter()

Process A can "publish" the greeter object by defining which methods are
exposed remotely and opening a Unix socket for incoming connections::

    factory = MethodCallServerFactory(greeter, ["hello"])
    reactor.listenUNIX("/some/socket/path", factory)

Then a second Python process "B" can connect to that socket and build a
"remote" greeter object, i.e. a proxy that forwards method calls to the
real greeter object living in process A::

    factory = MethodCallClientFactory()
    reactor.connectUNIX("/some/socket/path", factory)

    def got_remote(remote_greeter):
        deferred = remote_greeter.hello("Ted")
        deferred.addCallback(lambda result: ... # result == "hi Ted!")

    factory.getRemoteObject().addCallback(got_remote)

Note that when invoking a method via the remote proxy, the parameters
are required to be serializable with bpickle, so they can be sent over
the wire.

See also::

    http://twistedmatrix.com/documents/current/core/howto/amp.html

for more details about the Twisted AMP protocol.
"""
from uuid import uuid4

from twisted.internet.defer import Deferred, maybeDeferred, succeed
from twisted.internet.protocol import ServerFactory, ReconnectingClientFactory
from twisted.python.failure import Failure

from twisted.protocols.amp import (
    Argument, String, Integer, Command, AMP, MAX_VALUE_LENGTH, CommandLocator)

from landscape.lib.bpickle import loads, dumps, dumps_table


class MethodCallArgument(Argument):
    """A bpickle-compatible argument."""

    def toString(self, inObject):
        """Serialize an argument."""
        return dumps(inObject)

    def fromString(self, inString):
        """Unserialize an argument."""
        return loads(inString)

    @classmethod
    def check(cls, inObject):
        """Check if an argument is serializable."""
        return type(inObject) in dumps_table


class MethodCallError(Exception):
    """Raised when a L{MethodCall} command fails."""


class MethodCall(Command):
    """Call a method on the object exposed by a L{MethodCallServerFactory}.

    The command arguments have the following semantics:

    - C{sequence}: An integer uniquely indentifying a the L{MethodCall}
      being issued. The name 'sequence' is a bit misleading because it's
      really a uuid, since its values in practice are not in sequential
      order, they are just random values. The name is kept just for backward
      compatibility.

    - C{method}: The name of the method to invoke on the remote object.

    - C{arguments}: A BPickled binary tuple of the form C{(args, kwargs)},
      where C{args} are the positional arguments to be passed to the method
      and C{kwargs} the keyword ones.
    """

    arguments = [("sequence", Integer()),
                 ("method", String()),
                 ("arguments", String())]

    response = [("result", MethodCallArgument())]

    errors = {MethodCallError: "METHOD_CALL_ERROR"}


class MethodCallChunk(Command):
    """Send a chunk of L{MethodCall} containing a portion of the arguments.

    When a the arguments of a L{MethodCall} are bigger than 64k, they get split
    in several L{MethodCallChunk}s that are buffered on the receiver side.

    The command arguments have the following semantics:

    - C{sequence}: The unique integer associated with the L{MethodCall} that
      this L{MethodCallChunk} is part of.

    - C{chunk}: A portion of the big BPickle C{arguments} string which is
      being split and buffered.
    """

    arguments = [("sequence", Integer()),
                 ("chunk", String())]

    response = [("result", Integer())]

    errors = {MethodCallError: "METHOD_CALL_ERROR"}


class MethodCallReceiver(CommandLocator):
    """Expose methods of a local object over AMP.

    @param obj: The Python object to be exposed.
    @param methods: The list of the object's methods that can be called
         remotely.
    """

    def __init__(self, obj, methods):
        CommandLocator.__init__(self)
        self._object = obj
        self._methods = methods
        self._pending_chunks = {}

    @MethodCall.responder
    def receive_method_call(self, sequence, method, arguments):
        """Call an object's method with the given arguments.

        If a connected client sends a L{MethodCall} for method C{foo_bar}, then
        the actual method C{foo_bar} of the object associated with the protocol
        will be called with the given C{args} and C{kwargs} and its return
        value delivered back to the client as response to the command.

        @param sequence: The integer that uniquely identifies the L{MethodCall}
            being received.
        @param method: The name of the object's method to call.
        @param arguments: A bpickle'd binary tuple of (args, kwargs) to be
           passed to the method. In case this L{MethodCall} has been preceded
           by one or more L{MethodCallChunk}s, C{arguments} is the last chunk
           of data.
        """
        chunks = self._pending_chunks.pop(sequence, None)
        if chunks is not None:
            # We got some L{MethodCallChunk}s before, this is the last.
            chunks.append(arguments)
            arguments = "".join(chunks)

        args, kwargs = loads(arguments)

        if not method in self._methods:
            raise MethodCallError("Forbidden method '%s'" % method)

        method_func = getattr(self._object, method)

        def handle_result(result):
            return {"result": self._check_result(result)}

        def handle_failure(failure):
            raise MethodCallError(failure.value)

        deferred = maybeDeferred(method_func, *args, **kwargs)
        deferred.addCallback(handle_result)
        deferred.addErrback(handle_failure)
        return deferred

    @MethodCallChunk.responder
    def receive_method_call_chunk(self, sequence, chunk):
        """Receive a part of a multi-chunk L{MethodCall}.

        Add the received C{chunk} to the buffer of the L{MethodCall} identified
        by C{sequence}.
        """
        self._pending_chunks.setdefault(sequence, []).append(chunk)
        return {"result": sequence}

    def _check_result(self, result):
        """Check that the C{result} we're about to return is serializable.

        @return: The C{result} itself if valid.
        @raises: L{MethodCallError} if C{result} is not serializable.
        """
        if not MethodCallArgument.check(result):
            raise MethodCallError("Non-serializable result")
        return result


class MethodCallSender(object):
    """Call methods on a remote object over L{AMP} and return the result.

    @param protocol: A connected C{AMP} protocol.
    @param clock: An object implementing the C{IReactorTime} interface.

    @ivar timeout: A timeout for remote method class, see L{send_method_call}.
    """
    timeout = 60

    _chunk_size = MAX_VALUE_LENGTH

    def __init__(self, protocol, clock):
        self._protocol = protocol
        self._clock = clock

    def _call_remote_with_timeout(self, command, **kwargs):
        """Send an L{AMP} command that will errback in case of a timeout.

        @return: A deferred resulting in the command's response (or failure) if
            the peer responds within C{self.timeout} seconds, or that errbacks
            with a L{MethodCallError} otherwise.
        """
        deferred = Deferred()

        def handle_response(response):
            if not call.active():
                # Late response for a request that has timeout,
                # just ignore it.
                return
            call.cancel()
            deferred.callback(response)

        def handle_timeout():
            # The peer didn't respond on time, raise an error.
            deferred.errback(MethodCallError("timeout"))

        call = self._clock.callLater(self.timeout, handle_timeout)

        result = self._protocol.callRemote(command, **kwargs)
        result.addBoth(handle_response)
        return deferred

    def send_method_call(self, method, args=[], kwargs={}):
        """Send a L{MethodCall} command with the given arguments.

        If a response from the server is not received within C{self.timeout}
        seconds, the returned deferred will errback with a L{MethodCallError}.

        @param method: The name of the remote method to invoke.
        @param args: The positional arguments to pass to the remote method.
        @param kwargs: The keyword arguments to pass to the remote method.

        @return: A C{Deferred} firing with the return value of the method
            invoked on the remote object. If the remote method itself returns
            a deferred, we fire with the callback value of such deferred.
        """
        arguments = dumps((args, kwargs))
        sequence = uuid4().int

        # Split the given arguments in one or more chunks
        chunks = [arguments[i:i + self._chunk_size]
                  for i in xrange(0, len(arguments), self._chunk_size)]

        result = Deferred()
        if len(chunks) > 1:
            # If we have N chunks, send the first N-1 as MethodCallChunk's
            for chunk in chunks[:-1]:

                def create_send_chunk(sequence, chunk):
                    send_chunk = lambda x: self._protocol.callRemote(
                        MethodCallChunk, sequence=sequence, chunk=chunk)
                    return send_chunk

                result.addCallback(create_send_chunk(sequence, chunk))

        def send_last_chunk(ignored):
            chunk = chunks[-1]
            return self._call_remote_with_timeout(
                MethodCall, sequence=sequence, method=method, arguments=chunk)

        result.addCallback(send_last_chunk)
        result.addCallback(lambda response: response["result"])
        result.callback(None)
        return result


class MethodCallServerProtocol(AMP):
    """Receive L{MethodCall} commands over the wire and send back results."""

    def __init__(self, obj, methods):
        AMP.__init__(self, locator=MethodCallReceiver(obj, methods))


class MethodCallClientProtocol(AMP):
    """Send L{MethodCall} commands over the wire using the AMP protocol."""

    factory = None

    def connectionMade(self):
        """Notify our factory that we're ready to go."""
        if self.factory is not None:  # Factory can be None in unit-tests
            self.factory.clientConnectionMade(self)


class RemoteObject(object):
    """An object able to transparently call methods on a remote object.

    Any method call on a L{RemoteObject} instance will return a L{Deferred}
    resulting in the return value of the same method call performed on
    the remote object exposed by the peer.
    """

    def __init__(self, factory):
        """
        @param factory: The L{MethodCallClientFactory} used for connecting to
            the other peer. Look there if you need to tweak the behavior of
            this L{RemoteObject}.
        """
        self._sender = None
        self._pending_requests = {}
        self._factory = factory
        self._factory.notifyOnConnect(self._handle_connect)

    def __getattr__(self, method):
        """Return a function sending a L{MethodCall} for the given C{method}.

        When the created function is called, it sends the an appropriate
        L{MethodCall} to the remote peer passing it the arguments and
        keyword arguments it was called with, and returning a L{Deferred}
        resulting in the L{MethodCall}'s response value.
        """
        def send_method_call(*args, **kwargs):
            deferred = Deferred()
            self._send_method_call(method, args, kwargs, deferred)
            return deferred

        return send_method_call

    def _send_method_call(self, method, args, kwargs, deferred, call=None):
        """Send a L{MethodCall} command, adding callbacks to handle retries."""
        result = self._sender.send_method_call(method=method,
                                               args=args,
                                               kwargs=kwargs)
        result.addCallback(self._handle_result, deferred, call=call)
        result.addErrback(self._handle_failure, method, args, kwargs,
                          deferred, call=call)

        if self._factory.fake_connection is not None:
            # Transparently flush the connection after a send_method_call
            # invocation letting tests simulate a synchronous transport.
            # This is needed because the Twisted's AMP implementation
            # assume that the transport is asynchronous.
            self._factory.fake_connection.flush()

    def _handle_result(self, result, deferred, call=None):
        """Handles a successful C{send_method_call} result.

        @param response: The L{MethodCall} response.
        @param deferred: The deferred that was returned to the caller.
        @param call: If not C{None}, the scheduled timeout call associated with
            the given deferred.
        """
        if call is not None:
            call.cancel()  # This is a successful retry, cancel the timeout.
        deferred.callback(result)

    def _handle_failure(self, failure, method, args, kwargs, deferred,
                        call=None):
        """Called when a L{MethodCall} command fails.

        If a failure is due to a connection error and if C{retry_on_reconnect}
        is C{True}, we will try to perform the requested L{MethodCall} again
        as soon as a new connection becomes available, giving up after the
        specified C{timeout}, if any.

        @param failure: The L{Failure} raised by the requested L{MethodCall}.
        @param name: The method name associated with the failed L{MethodCall}.
        @param args: The positional arguments of the failed L{MethodCall}.
        @param kwargs: The keyword arguments of the failed L{MethodCall}.
        @param deferred: The deferred that was returned to the caller.
        @param call: If not C{None}, the scheduled timeout call associated with
            the given deferred.
        """
        is_method_call_error = failure.type is MethodCallError
        dont_retry = self._factory.retryOnReconnect is False

        if is_method_call_error or dont_retry:
            # This means either that the connection is working, and a
            # MethodCall protocol error occured, or that we gave up
            # trying and raised a timeout. In any case just propagate
            # the error.
            if deferred in self._pending_requests:
                self._pending_requests.pop(deferred)
            if call:
                call.cancel()
            deferred.errback(failure)
            return

        if self._factory.retryTimeout and call is None:
            # This is the first failure for this request, let's schedule a
            # timeout call.
            timeout = Failure(MethodCallError("timeout"))
            call = self._factory.clock.callLater(self._factory.retryTimeout,
                                                 self._handle_failure,
                                                 timeout, method, args,
                                                 kwargs, deferred=deferred)

        self._pending_requests[deferred] = (method, args, kwargs, call)

    def _handle_connect(self, protocol):
        """Handles a reconnection.

        @param protocol: The newly connected protocol instance.
        """
        self._sender = MethodCallSender(protocol, self._factory.clock)
        if self._factory.retryOnReconnect:
            self._retry()

    def _retry(self):
        """Try to perform again requests that failed."""

        # We need to copy the requests list before iterating over it, because
        # if we are actually still disconnected, callRemote will return a
        # failed deferred and the _handle_failure errback will be executed
        # synchronously during the loop, modifing the requests list itself.
        requests = self._pending_requests.copy()
        self._pending_requests.clear()

        while requests:
            deferred, (method, args, kwargs, call) = requests.popitem()
            self._send_method_call(method, args, kwargs, deferred, call=call)


class MethodCallServerFactory(ServerFactory):
    """Expose a Python object using L{MethodCall} commands over C{AMP}."""

    protocol = MethodCallServerProtocol

    def __init__(self, obj, methods):
        """
        @param object: The object exposed by the L{MethodCallProtocol}s
            instances created by this factory.
        @param methods: A list of the names of the methods that remote peers
            are allowed to call on the C{object} that we publish.
        """
        self.object = obj
        self.methods = methods

    def buildProtocol(self, addr):
        protocol = self.protocol(self.object, self.methods)
        protocol.factory = self
        return protocol


class MethodCallClientFactory(ReconnectingClientFactory):
    """
    Factory for L{MethodCallClientProtocol}s exposing an object or connecting
    to L{MethodCall} servers.

    When used to connect, if the connection fails or is lost the factory
    will keep retrying to establish it.

    @ivar factor: The time factor by which the delay between two subsequent
        connection retries will increase.
    @ivar maxDelay: Maximum number of seconds between connection attempts.
    @ivar protocol: The factory used to build protocol instances.
    @ivar remote: The factory used to build remote object instances.
    @ivar retryOnReconnect: If C{True}, the remote object returned by the
        C{getRemoteObject} method will retry requests that failed, as a
        result of a lost connection, as soon as a new connection is available.
    @param retryTimeout: A timeout for retrying requests, if the remote object
        can't perform them again successfully within this number of seconds,
        they will errback with a L{MethodCallError}.
    """

    factor = 1.6180339887498948
    maxDelay = 30

    protocol = MethodCallClientProtocol
    remote = RemoteObject

    retryOnReconnect = False
    retryTimeout = None

    # XXX support exposing fake asynchronous connections created by tests, so
    # they can be flushed transparently and emulate a synchronous behavior. See
    # also http://twistedmatrix.com/trac/ticket/6502, once that's fixed this
    # hack can be removed.
    fake_connection = None

    def __init__(self, clock):
        """
        @param object: The object exposed by the L{MethodCallProtocol}s
            instances created by this factory.
        @param reactor: The reactor used by the created protocols
            to schedule notifications and timeouts.
        """
        self.clock = clock
        self.delay = self.initialDelay
        self._connects = []
        self._requests = []
        self._remote = None

    def getRemoteObject(self):
        """Get a L{RemoteObject} as soon as the connection is ready.

        @return: A C{Deferred} firing with a connected L{RemoteObject}.
        """
        if self._remote is not None:
            return succeed(self._remote)
        deferred = Deferred()
        self._requests.append(deferred)
        return deferred

    def notifyOnConnect(self, callback):
        """Invoke the given C{callback} when a connection is re-established."""
        self._connects.append(callback)

    def dontNotifyOnConnect(self, callback):
        """Remove the given C{callback} from listeners."""
        self._connects.remove(callback)

    def clientConnectionMade(self, protocol):
        """Called when a newly built protocol gets connected."""
        if self._remote is None:
            # This is the first time we successfully connect
            self._remote = self.remote(self)

        for callback in self._connects:
            callback(protocol)

        # In all cases fire pending requests
        self._fire_requests(self._remote)

    def clientConnectionFailed(self, connector, reason):
        """Try to connect again or errback pending request."""
        ReconnectingClientFactory.clientConnectionFailed(self, connector,
                                                         reason)
        if self._callID is None:
            # The factory won't retry to connect, so notify that we failed
            self._fire_requests(reason)

    def buildProtocol(self, addr):
        self.resetDelay()
        protocol = ReconnectingClientFactory.buildProtocol(self, addr)
        return protocol

    def _fire_requests(self, result):
        """
        Fire all pending L{getRemoteObject} deferreds with the given C{result}.
        """
        requests = self._requests[:]
        self._requests = []

        for deferred in requests:
            deferred.callback(result)



from landscape.lib.hashlib import md5

def build_script(interpreter, code):
    """
    Concatenates a interpreter and script into an executable script.
    """
    return "#!%s\n%s" % ((interpreter or u"").encode("utf-8"),
                         (code or u"").encode("utf-8"))


def generate_script_hash(script):
    """
    Return a hash for a given script.
    """
    return md5(script).hexdigest()



from twisted.internet.defer import succeed, fail

from configobj import ConfigObj

from landscape.manager.manager import FAILED
from landscape.tests.helpers import LandscapeTest, DEFAULT_ACCEPTED_TYPES
from landscape.broker.tests.helpers import (
    BrokerServerHelper, RemoteClientHelper)
from landscape.broker.tests.test_ping import FakePageGetter


class FakeClient(object):
    pass


class FakeCreator(object):

    def __init__(self, reactor, config):
        pass

    def connect(self):
        return succeed(FakeClient())


class BrokerServerTest(LandscapeTest):

    helpers = [BrokerServerHelper]

    def test_ping(self):
        """
        The L{BrokerServer.ping} simply returns C{True}.
        """
        self.assertTrue(self.broker.ping())

    def test_get_session_id(self):
        """
        The L{BrokerServer.get_session_id} method gets the same
        session ID from the L{MessageStore} until it is dropped.
        """
        session_id1 = self.broker.get_session_id()
        session_id2 = self.broker.get_session_id()
        self.assertEqual(session_id1, session_id2)
        self.mstore.drop_session_ids()
        session_id3 = self.broker.get_session_id()
        self.assertNotEqual(session_id1, session_id3)

    def test_get_session_id_with_scope(self):
        """
        The L{BrokerServer.get_session_id} method gets the same session ID from
        the L{MessageStore} for the same scope, but a new session ID for a new
        scope.
        """
        disk_session_id1 = self.broker.get_session_id(scope="disk")
        disk_session_id2 = self.broker.get_session_id(scope="disk")
        users_session_id = self.broker.get_session_id(scope="users")
        self.assertEqual(disk_session_id1, disk_session_id2)
        self.assertNotEqual(disk_session_id1, users_session_id)

    def test_send_message(self):

        """
        The L{BrokerServer.send_message} method forwards a message to the
        broker's exchanger.
        """
        message = {"type": "test"}
        self.mstore.set_accepted_types(["test"])
        session_id = self.broker.get_session_id()
        self.broker.send_message(message, session_id)
        self.assertMessages(self.mstore.get_pending_messages(), [message])
        self.assertFalse(self.exchanger.is_urgent())

    def test_send_message_with_urgent(self):
        """
        The L{BrokerServer.send_message} can optionally specify the urgency
        of the message.
        """
        message = {"type": "test"}
        self.mstore.set_accepted_types(["test"])
        session_id = self.broker.get_session_id()
        self.broker.send_message(message, session_id, urgent=True)
        self.assertMessages(self.mstore.get_pending_messages(), [message])
        self.assertTrue(self.exchanger.is_urgent())

    def test_send_message_wont_send_with_invalid_session_id(self):
        """
        The L{BrokerServer.send_message} call will silently drop messages
        that have invalid session ids as they must have been generated
        prior to the last resync request - this guards against out of
        context data being sent to the server.
        """
        message = {"type": "test"}
        self.mstore.set_accepted_types(["test"])
        self.broker.send_message(message, "Not Valid")
        self.assertMessages(self.mstore.get_pending_messages(), [])

    def test_send_message_with_none_as_session_id_raises(self):
        """
        We should never call C{send_message} without first obtaining a session
        id.  Attempts to do so should raise to alert the developer to their
        mistake.
        """
        message = {"type": "test"}
        self.mstore.set_accepted_types(["test"])
        self.assertRaises(
            RuntimeError, self.broker.send_message, message, None)

    def test_is_pending(self):
        """
        The L{BrokerServer.is_pending} method indicates if a message with
        the given id is pending waiting for delivery in the message store.
        """
        self.assertFalse(self.broker.is_message_pending(123))
        message = {"type": "test"}
        self.mstore.set_accepted_types(["test"])
        session_id = self.broker.get_session_id()
        message_id = self.broker.send_message(message, session_id)
        self.assertTrue(self.broker.is_message_pending(message_id))

    def test_register_client(self):
        """
        The L{BrokerServer.register_client} method can be used to register
        client components that need to communicate with the server. After
        the registration they can be fetched with L{BrokerServer.get_clients}.
        """
        self.assertEqual(self.broker.get_clients(), [])
        self.assertEqual(self.broker.get_client("test"), None)
        self.assertEqual(self.broker.get_connectors(), [])
        self.assertEqual(self.broker.get_connector("test"), None)

        def assert_registered(ignored):
            self.assertEqual(len(self.broker.get_clients()), 1)
            self.assertEqual(len(self.broker.get_connectors()), 1)
            self.assertTrue(
                isinstance(self.broker.get_client("test"), FakeClient))
            self.assertTrue(
                isinstance(self.broker.get_connector("test"), FakeCreator))

        self.broker.connectors_registry = {"test": FakeCreator}
        result = self.broker.register_client("test")
        return result.addCallback(assert_registered)

    def test_stop_clients(self):
        """
        The L{BrokerServer.stop_clients} method calls the C{exit} method
        of each registered client, and returns a deferred resulting in C{None}
        if all C{exit} calls were successful.
        """
        self.broker.connectors_registry = {"foo": FakeCreator,
                                           "bar": FakeCreator}
        self.broker.register_client("foo")
        self.broker.register_client("bar")
        for client in self.broker.get_clients():
            client.exit = self.mocker.mock()
            self.expect(client.exit()).result(succeed(None))
        self.mocker.replay()
        return self.assertSuccess(self.broker.stop_clients())

    def test_stop_clients_with_failure(self):
        """
        The L{BrokerServer.stop_clients} method calls the C{exit} method
        of each registered client, and returns a deferred resulting in C{None}
        if all C{exit} calls were successful.
        """
        self.broker.connectors_registry = {"foo": FakeCreator,
                                           "bar": FakeCreator}
        self.broker.register_client("foo")
        self.broker.register_client("bar")
        [client1, client2] = self.broker.get_clients()
        client1.exit = self.mocker.mock()
        client2.exit = self.mocker.mock()
        self.expect(client1.exit()).result(succeed(None))
        self.expect(client2.exit()).result(fail(Exception()))
        self.mocker.replay()
        return self.assertFailure(self.broker.stop_clients(), Exception)

    def test_reload_configuration(self):
        """
        The L{BrokerServer.reload_configuration} method forces the config
        file associated with the broker server to be reloaded.
        """
        config_obj = ConfigObj(self.config_filename)
        config_obj["client"]["computer_title"] = "New Title"
        config_obj.write()
        result = self.broker.reload_configuration()
        result.addCallback(lambda x: self.assertEqual(
            self.config.computer_title, "New Title"))
        return result

    def test_reload_configuration_stops_clients(self):
        """
        The L{BrokerServer.reload_configuration} method forces the config
        file associated with the broker server to be reloaded.
        """
        self.broker.connectors_registry = {"foo": FakeCreator,
                                           "bar": FakeCreator}
        self.broker.register_client("foo")
        self.broker.register_client("bar")
        for client in self.broker.get_clients():
            client.exit = self.mocker.mock()
            self.expect(client.exit()).result(succeed(None))
        self.mocker.replay()
        return self.assertSuccess(self.broker.reload_configuration())

    def test_register(self):
        """
        The L{BrokerServer.register} method attempts to register with the
        Landscape server and waits for a C{set-id} message from it.
        """
        registered = self.broker.register()
        # This should callback the deferred.
        self.exchanger.handle_message({"type": "set-id", "id": "abc",
                                       "insecure-id": "def"})
        return self.assertSuccess(registered)

    def test_get_accepted_types_empty(self):
        """
        The L{BrokerServer.get_accepted_message_types} returns an empty list
        if no message types are accepted by the Landscape server.
        """
        self.mstore.set_accepted_types([])
        self.assertEqual(self.broker.get_accepted_message_types(), [])

    def test_get_accepted_message_types(self):
        """
        The L{BrokerServer.get_accepted_message_types} returns the list of
        message types accepted by the Landscape server.
        """
        self.mstore.set_accepted_types(["foo", "bar"])
        self.assertEqual(sorted(self.broker.get_accepted_message_types()),
                         ["bar", "foo"])

    def test_get_server_uuid_with_unset_uuid(self):
        """
        The L{BrokerServer.get_server_uuid} method returns C{None} if the uuid
        of the Landscape server we're pointing at is unknown.
        """
        self.assertEqual(self.broker.get_server_uuid(), None)

    def test_get_server_uuid(self):
        """
        The L{BrokerServer.get_server_uuid} method returns the uuid of the
        Landscape server we're pointing at.
        """
        self.mstore.set_server_uuid("the-uuid")
        self.assertEqual(self.broker.get_server_uuid(), "the-uuid")

    def test_register_client_accepted_message_type(self):
        """
        The L{BrokerServer.register_client_accepted_message_type} method can
        register new message types accepted by this Landscape client.
        """
        self.broker.register_client_accepted_message_type("type1")
        self.broker.register_client_accepted_message_type("type2")
        self.assertEqual(self.exchanger.get_client_accepted_message_types(),
                         sorted(["type1", "type2"] + DEFAULT_ACCEPTED_TYPES))

    def test_fire_event(self):
        """
        The L{BrokerServer.fire_event} method fires an event in the broker
        reactor.
        """
        callback = self.mocker.mock()
        callback()
        self.mocker.replay()
        self.reactor.call_on("event", callback)
        self.broker.fire_event("event")

    def test_exit(self):
        """
        The L{BrokerServer.exit} method stops all registered clients.
        """
        self.broker.connectors_registry = {"foo": FakeCreator,
                                           "bar": FakeCreator}
        self.broker.register_client("foo")
        self.broker.register_client("bar")
        for client in self.broker.get_clients():
            client.exit = self.mocker.mock()
            self.expect(client.exit()).result(succeed(None))
        self.mocker.replay()
        return self.assertSuccess(self.broker.exit())

    def test_exit_exits_when_other_daemons_blow_up(self):
        """
        If a broker client blow up in its exit() methods, exit should ignore
        the error and exit anyway.
        """
        self.broker.connectors_registry = {"foo": FakeCreator}
        self.broker.register_client("foo")
        [client] = self.broker.get_clients()
        client.exit = self.mocker.mock()
        self.expect(client.exit()).result(fail(ZeroDivisionError()))
        self.mocker.replay()

        def assert_event(ignored):
            self.reactor.advance(1)

        result = self.broker.exit()
        return result.addCallback(assert_event)

    def test_exit_fires_reactor_events(self):
        """
        The L{BrokerServer.exit} method stops the reactor after having
        requested all broker clients to shutdown.
        """
        self.broker.connectors_registry = {"foo": FakeCreator}
        self.broker.register_client("foo")
        [client] = self.broker.get_clients()
        self.mocker.order()

        client.exit = self.mocker.mock()
        self.reactor.stop = self.mocker.mock()
        self.broker.stop_exchanger()
        self.expect(client.exit()).result(fail(ZeroDivisionError()))
        self.reactor.stop()
        self.mocker.replay()

        def assert_stopped(ignored):
            self.reactor.advance(1)

        result = self.broker.exit()
        return result.addCallback(assert_stopped)

    def test_listen_events(self):
        """
        The L{BrokerServer.listen_events} method returns a deferred which is
        fired when the first of the given events occurs.
        """
        result = self.broker.listen_events(["event1", "event2"])
        self.reactor.fire("event2")
        return self.assertSuccess(result, "event2")

    def test_listen_event_only_once(self):
        """
        The L{BrokerServer.listen_events} listens only to one occurrence of
        the given events.
        """
        result = self.broker.listen_events(["event"])
        self.assertEqual(self.reactor.fire("event"), [None])
        self.assertEqual(self.reactor.fire("event"), [])
        return self.assertSuccess(result, "event")

    def test_listen_events_call_cancellation(self):
        """
        The L{BrokerServer.listen_events} cleanly cancels event calls for
        unfired events, without interfering with unrelated handlers.
        """
        self.broker.listen_events(["event"])
        self.reactor.call_on("event", lambda: 123)  # Unrelated handler
        self.assertEqual(self.reactor.fire("event"), [None, 123])

    def test_stop_exchanger(self):
        """
        The L{BrokerServer.stop_exchanger} stops the exchanger so no further
        messages are sent or consumed.
        """
        self.pinger.start()
        self.exchanger.schedule_exchange()
        self.broker.stop_exchanger()

        self.reactor.advance(self.config.exchange_interval)
        self.assertFalse(self.transport.payloads)

    def test_stop_exchanger_stops_pinger(self):
        """
        The L{BrokerServer.stop_exchanger} stops the pinger and no further
        pings are performed.
        """
        url = "http://example.com/mysuperping"
        page_getter = FakePageGetter(None)
        self.pinger.start()
        self.config.ping_url = url
        self.pinger._ping_client.get_page = page_getter.get_page
        self.identity.insecure_id = 23

        self.broker.stop_exchanger()
        self.reactor.advance(self.config.exchange_interval)
        self.assertEqual([], page_getter.fetches)


class EventTest(LandscapeTest):

    helpers = [RemoteClientHelper]

    def test_resynchronize(self):
        """
        The L{BrokerServer.resynchronize} method broadcasts a C{resynchronize}
        event to all connected clients.
        """
        callback = self.mocker.mock()
        self.expect(callback(["foo"])).result("foo")
        self.mocker.replay()
        self.client_reactor.call_on("resynchronize", callback)
        return self.assertSuccess(self.broker.resynchronize(["foo"]),
                                  [["foo"]])

    def test_impending_exchange(self):
        """
        The L{BrokerServer.impending_exchange} method broadcasts an
        C{impending-exchange} event to all connected clients.
        """
        plugin = self.mocker.mock()
        plugin.register(self.client)
        plugin.exchange()
        self.mocker.replay()
        self.client.add(plugin)
        return self.assertSuccess(self.broker.impending_exchange(), [[None]])

    def test_broker_started(self):
        """
        The L{BrokerServer.broker_started} method broadcasts a
        C{broker-started} event to all connected clients, which makes them
        re-registered any previously registered accepted message type.
        """

        def assert_broker_started(ignored):
            self.remote.register_client_accepted_message_type = \
                                                        self.mocker.mock()
            self.remote.register_client_accepted_message_type("type")
            self.remote.register_client = self.mocker.mock()
            self.remote.register_client("client")
            self.mocker.replay()
            return self.assertSuccess(self.broker.broker_reconnect(), [[None]])

        registered = self.client.register_message("type", lambda x: None)
        return registered.addCallback(assert_broker_started)

    def test_server_uuid_changed(self):
        """
        The L{BrokerServer.server_uuid_changed} method broadcasts a
        C{server-uuid-changed} event to all connected clients.
        """
        callback = self.mocker.mock()
        callback(None, "abc")
        self.mocker.replay()
        self.client_reactor.call_on("server-uuid-changed", callback)
        return self.assertSuccess(self.broker.server_uuid_changed(None, "abc"),
                                  [[None]])

    def test_message_type_acceptance_changed(self):
        """
        The L{BrokerServer.message_type_acceptance_changed} method broadcasts
        a C{message-type-acceptance-changed} event to all connected clients.
        """
        callback = self.mocker.mock()
        callback(True)
        self.mocker.replay()
        self.client_reactor.call_on(
            ("message-type-acceptance-changed", "type"), callback)
        result = self.broker.message_type_acceptance_changed("type", True)
        return self.assertSuccess(result, [[None]])

    def test_package_data_changed(self):
        """
        The L{BrokerServer.package_data_changed} method broadcasts a
        C{package-data-changed} event to all connected clients.
        """
        callback = self.mocker.mock()
        callback()
        self.mocker.replay()
        self.client_reactor.call_on("package-data-changed", callback)
        return self.assertSuccess(self.broker.package_data_changed(), [[None]])


class HandlersTest(LandscapeTest):

    helpers = [BrokerServerHelper]

    def setUp(self):
        super(HandlersTest, self).setUp()
        self.broker.connectors_registry = {"test": FakeCreator}
        self.broker.register_client("test")
        self.client = self.broker.get_client("test")

    def test_message(self):
        """
        The L{BrokerServer} calls the C{message} method on all
        registered plugins when messages are received from the server.
        """
        message = {"type": "foobar", "value": 42}
        self.client.message = self.mocker.mock()
        self.client.message(message)
        self.mocker.result(succeed(True))
        self.mocker.replay()
        self.transport.responses.append([{"type": "foobar", "value": 42}])
        self.exchanger.exchange()

    def test_message_failed_operation_without_plugins(self):
        """
        When there are no broker plugins available to handle a message, an
        operation-result message should be sent back to the server indicating a
        failure.
        """
        self.log_helper.ignore_errors("Nobody handled the foobar message.")
        self.mstore.set_accepted_types(["operation-result"])
        message = {"type": "foobar", "operation-id": 4}
        self.client.message = self.mocker.mock()
        self.client.message(message)
        self.mocker.result(succeed(False))
        self.mocker.replay()
        result = self.reactor.fire("message", message)
        result = [i for i in result if i is not None][0]

        class StartsWith(object):

            def __eq__(self, other):
                return other.startswith(
                    "Landscape client failed to handle this request (foobar)")

        def broadcasted(ignored):
            self.assertMessages(
                self.mstore.get_pending_messages(),
                [{"type": "operation-result", "status": FAILED,
                  "result-text": StartsWith(), "operation-id": 4}])

        result.addCallback(broadcasted)
        return result

    def test_impending_exchange(self):
        """
        When an C{impending-exchange} event is fired by the reactor, the
        broker broadcasts it to its clients.
        """
        self.client.fire_event = self.mocker.mock()
        self.client.fire_event("impending-exchange")
        self.mocker.result(succeed(None))
        self.mocker.replay()
        self.reactor.fire("impending-exchange")

    def test_message_type_acceptance_changed(self):
        """
        When a C{message-type-acceptance-changed} event is fired by the
        reactor, the broker broadcasts it to its clients.
        """
        self.client.fire_event = self.mocker.mock()
        self.client.fire_event("message-type-acceptance-changed", "test", True)
        self.mocker.result(succeed(None))
        self.mocker.replay()
        self.reactor.fire("message-type-acceptance-changed", "test", True)

    def test_server_uuid_changed(self):
        """
        When a C{server-uuid-changed} event is fired by the reactor, the
        broker broadcasts it to its clients.
        """
        self.client.fire_event = self.mocker.mock()
        self.client.fire_event("server-uuid-changed", None, 123)
        self.mocker.result(succeed(None))
        self.mocker.replay()
        self.reactor.fire("server-uuid-changed", None, 123)

    def test_package_data_changed(self):
        """
        When a C{package-data-changed} event is fired by the reactor, the
        broker broadcasts it to its clients.
        """
        self.client.fire_event = self.mocker.mock()
        self.client.fire_event("package-data-changed")
        self.mocker.result(succeed(None))
        self.mocker.replay()
        self.reactor.fire("package-data-changed")

    def test_resynchronize_clients(self):
        """
        When a C{resynchronize} event is fired by the reactor, the
        broker broadcasts it to its clients.
        """
        self.client.fire_event = self.mocker.mock()
        self.client.fire_event("resynchronize")
        self.mocker.result(succeed(None))
        self.mocker.replay()
        self.reactor.fire("resynchronize-clients")



from landscape.sysinfo.sysinfo import SysInfoPluginRegistry
from landscape.sysinfo.network import Network
from landscape.tests.helpers import LandscapeTest


class NetworkTest(LandscapeTest):

    def setUp(self):
        super(NetworkTest, self).setUp()
        self.result = []
        self.network = Network(lambda: self.result)
        self.sysinfo = SysInfoPluginRegistry()
        self.sysinfo.add(self.network)

    def test_run_returns_succeeded_deferred(self):
        """L{Network.run} always returns a succeeded C{Deferred}."""
        self.assertIs(None, self.successResultOf(self.network.run()))

    def test_run_adds_header(self):
        """
        A header is written to sysinfo output for each network device reported
        by L{get_active_device_info}.
        """
        self.result = [{"interface": "eth0", "ip_address": "192.168.0.50"}]
        self.network.run()
        self.assertEqual([("IP address for eth0", "192.168.0.50")],
                          self.sysinfo.get_headers())

    def test_run_without_network_devices(self):
        """
        If no network device information is available, no headers are added to
        the sysinfo output.
        """
        self.network.run()
        self.assertEqual([], self.sysinfo.get_headers())



from landscape.configuration import LandscapeSetupConfiguration
from landscape.tests.helpers import LandscapeTest
from landscape.ui.tests.helpers import (
    dbus_test_should_skip, dbus_skip_message)
if not dbus_test_should_skip:
    import dbus
    from landscape.ui.model.configuration.mechanism import (
        ConfigurationMechanism, INTERFACE_NAME)


class MechanismTest(LandscapeTest):
    """
    Test that we can use mechanism calls successfully from within a secure
    context (the easiest to achieve is in-process calls.
    """

    def setUp(self):
        super(MechanismTest, self).setUp()
        config = "[client]\n"
        config += "data_path = /var/lib/landscape/client/\n"
        config += "http_proxy = http://proxy.localdomain:3192\n"
        config += "tags = a_tag\n"
        config += "url = https://landscape.canonical.com/message-system\n"
        config += "account_name = foo\n"
        config += "registration_key = boink\n"
        config += "computer_title = baz\n"
        config += "https_proxy = https://proxy.localdomain:6192\n"
        config += "ping_url = http://landscape.canonical.com/ping\n"
        self.config_filename = self.makeFile(config)
        self.config = LandscapeSetupConfiguration()
        dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
        bus = dbus.SessionBus(private=True)
        bus_name = dbus.service.BusName(INTERFACE_NAME, bus)
        self.mechanism = ConfigurationMechanism(self.config, bus_name)
        self.config.load(["-c", self.config_filename])

    def tearDown(self):
        self.mechanism.remove_from_connection()
        super(MechanismTest, self).tearDown()

    def test_is_local_call(self):
        """
        Test simple mechanism for checking if a call is local does the right
        thing.  Anything passed to this function that is not L{None} will
        result in is returning False - this in turn means that bypassing
        security will not happen, which is the right thing in failure cases
        too.
        """
        self.assertTrue(self.mechanism._is_local_call(None, None))
        self.assertFalse(self.mechanism._is_local_call(True, True))

    def test_get_account_name(self):
        """
        Test we can get account name from the mechanism.
        """
        self.assertEqual("foo", self.mechanism.get("account_name"))

    def test_set_account_name(self):
        """
        Test we can set the account name via the mechanism.
        """
        self.mechanism.set("account_name", "bar")
        self.assertEqual("bar", self.mechanism.get("account_name"))

    def test_set_account_name_unicode(self):
        """
        Non-ascii characters are replaced before passing to underlying config.
        """
        self.mechanism.set("account_name", u"unicode\u00a3unicode")
        self.assertEqual("unicode?unicode", self.mechanism.get("account_name"))

    def test_no_unicode_to_underlying_config(self):
        """
        Non-ascii characters are replaced before passing to underlying config.
        """
        class FakeConfig(object):
            def __init__(self):
                self.account_name = None

        fake_config = FakeConfig()
        self.mechanism._config = fake_config
        self.mechanism.set("account_name", u"unicode\u00a3unicode")
        self.assertEqual("unicode?unicode", fake_config.account_name)

    def test_get_data_path(self):
        """
        Test we can get the data path from the mechanism.
        """
        self.assertEqual("/var/lib/landscape/client/",
                         self.mechanism.get("data_path"))

    def set_data_path(self):
        """
        Test we can set the data path via the mechanism.
        """
        self.mechanism.set("data_path", "bar")
        self.assertEqual("bar", self.mechanism.get("data_path"))

    def test_get_http_proxy(self):
        """
        Test that we can get the HTTP proxy from the mechanism.
        """
        self.assertEqual("http://proxy.localdomain:3192",
                         self.mechanism.get("http_proxy"))

    def test_set_http_proxy(self):
        """
        Test that we can set the HTTP proxy via the mechanism.
        """
        self.mechanism.set("http_proxy", "bar")
        self.assertEqual("bar", self.mechanism.get("http_proxy"))

    def test_get_tags(self):
        """
        Test that we can get Tags from the mechanism.
        """
        self.assertEqual("a_tag", self.mechanism.get("tags"))

    def test_set_tags(self):
        """
        Test that we can set Tags via the mechanism.
        """
        self.mechanism.set("tags", "bar")
        self.assertEqual("bar", self.mechanism.get("tags"))

    def test_get_url(self):
        """
        Test that we can get URL from the mechanism.
        """
        self.assertEqual("https://landscape.canonical.com/message-system",
                          self.mechanism.get("url"))

    def test_set_url(self):
        """
        Test that we can set the URL via the mechanisms.
        """
        self.mechanism.set("url", "bar")
        self.assertEqual(self.mechanism.get("url"), "bar")

    def test_get_ping_url(self):
        """
        Test that we can get the Ping URL from the mechanism.
        """
        self.assertEqual("http://landscape.canonical.com/ping",
                          self.mechanism.get("ping_url"))

    def test_set_ping_url(self):
        """
        Test that we can set the Ping URL via the mechanism.
        """
        self.mechanism.set("ping_url", "bar")
        self.assertEqual("bar", self.mechanism.get("ping_url"))

    def test_get_registration_key(self):
        """
        Test that we can get the registration key from the mechanism.
        """
        self.assertEqual("boink", self.mechanism.get("registration_key"))

    def test_set_registration_key(self):
        """
        Test that we can set the registration key via the mechanism.
        """
        self.mechanism.set("registration_key", "bar")
        self.assertEqual("bar", self.mechanism.get("registration_key"))

    def test_get_computer_title(self):
        """
        Test that we can get the computer title from the mechanism.
        """
        self.assertEqual("baz", self.mechanism.get("computer_title"))

    def test_set_computer_title(self):
        """
        Test that we can set the computer title via the mechanism.
        """
        self.mechanism.set("computer_title", "bar")
        self.assertEqual("bar", self.mechanism.get("computer_title"))

    def test_get_https_proxy(self):
        """
        Test that we can get the HTTPS Proxy from the mechanism.
        """
        self.assertEqual("https://proxy.localdomain:6192",
                         self.mechanism.get("https_proxy"))

    def test_set_https_proxy(self):
        """
        Test that we can set the HTTPS Proxy via the mechanism.
        """
        self.mechanism.set("https_proxy", "bar")
        self.assertEqual("bar", self.mechanism.get("https_proxy"))

    def test_exit(self):
        """
        Test that we cause the mechanism to exit.
        """
        self.assertRaises(SystemExit, self.mechanism.exit)

    if dbus_test_should_skip:
        skip = dbus_skip_message



DEBIAN_REVISION = ""
UPSTREAM_VERSION = "13.10+bzr73"
VERSION = "%s%s" % (UPSTREAM_VERSION, DEBIAN_REVISION)

# The "server-api" field of outgoing messages will be set to this value, and
# used by the server message system to lookup the correct MessageAPI adapter
# for handling the messages sent by the client. Bump it when the schema of any
# of the messages sent by the client changes in a backward-incompatible way.
#
# Changelog:
#
# 3.2:
#  * Add new "eucalyptus-info" and "eucalyptus-info-error" messages.
#
SERVER_API = "3.2"

# XXX This is needed for backward compatibility in the server code importing
# the API variable. We should eventually replace it in the server code.
API = SERVER_API

# The "client-api" field of outgoing messages will be set to this value, and
# used by the server to know which schema do the message types accepted by the
# client support. Bump it when the schema of an accepted message type changes
# and update the changelog below as needed.
#
# Changelog:
#
# 3.3:
#  * Add "binaries" field to "change-packages"
#  * Add "policy" field to "change-packages"
#  * Add new "change-package-locks" client accepted message type.
#
# 3.4:
#  * Add "hold" field to "change-packages"
#  * Add "remove-hold" field to "change-packages"
#
# 3.5:
#  * Support per-exchange authentication tokens
#
# 3.6:
#  * Handle scopes in resynchronize requests
#
# 3.7:
#  * Server returns 402 Payment Required if the computer has no valid license.
#
CLIENT_API = "3.7"



import os

from landscape import VERSION
from landscape.broker.transport import HTTPTransport
from landscape.lib.fetch import PyCurlError
from landscape.lib import bpickle

from landscape.tests.helpers import LandscapeTest, LogKeeperHelper

from twisted.web import server, resource
from twisted.internet import reactor
from twisted.internet.ssl import DefaultOpenSSLContextFactory
from twisted.internet.threads import deferToThread


def sibpath(path):
    return os.path.join(os.path.dirname(__file__), path)


PRIVKEY = sibpath("private.ssl")
PUBKEY = sibpath("public.ssl")
BADPRIVKEY = sibpath("badprivate.ssl")
BADPUBKEY = sibpath("badpublic.ssl")


class DataCollectingResource(resource.Resource):

    request = content = None

    def getChild(self, request, name):
        return self

    def render(self, request):
        self.request = request
        self.content = request.content.read()
        return bpickle.dumps("Great.")


class HTTPTransportTest(LandscapeTest):

    helpers = [LogKeeperHelper]

    def setUp(self):
        super(HTTPTransportTest, self).setUp()
        self.ports = []

    def tearDown(self):
        super(HTTPTransportTest, self).tearDown()
        for port in self.ports:
            port.stopListening()

    def test_get_url(self):
        url = "http://example/ooga"
        transport = HTTPTransport(None, url)
        self.assertEqual(transport.get_url(), url)

    def test_set_url(self):
        transport = HTTPTransport(None, "http://example/ooga")
        transport.set_url("http://example/message-system")
        self.assertEqual(transport.get_url(), "http://example/message-system")

    def test_request_data(self):
        """
        When a request is sent with HTTPTransport.exchange, it should
        include the (optional) computer ID, a user agent, and the
        message API version as HTTP headers, and the payload as a
        bpickled request body.
        """
        r = DataCollectingResource()
        port = reactor.listenTCP(0, server.Site(r), interface="127.0.0.1")
        self.ports.append(port)
        transport = HTTPTransport(
            None, "http://localhost:%d/" % (port.getHost().port,))
        result = deferToThread(transport.exchange, "HI", computer_id="34",
                               exchange_token="abcd-efgh", message_api="X.Y")

        def got_result(ignored):
            self.assertEqual(r.request.received_headers["x-computer-id"],
                             "34")
            self.assertEqual(r.request.received_headers["x-exchange-token"],
                             "abcd-efgh")
            self.assertEqual(r.request.received_headers["user-agent"],
                             "landscape-client/%s" % (VERSION,))
            self.assertEqual(r.request.received_headers["x-message-api"],
                             "X.Y")
            self.assertEqual(bpickle.loads(r.content), "HI")
        result.addCallback(got_result)
        return result

    def test_ssl_verification_positive(self):
        """
        The client transport should complete an upload of messages to
        a host which provides SSL data which can be verified by the
        public key specified.
        """
        r = DataCollectingResource()
        context_factory = DefaultOpenSSLContextFactory(PRIVKEY, PUBKEY)
        port = reactor.listenSSL(0, server.Site(r), context_factory,
                                 interface="127.0.0.1")
        self.ports.append(port)
        transport = HTTPTransport(
            None, "https://localhost:%d/" % (port.getHost().port,), PUBKEY)
        result = deferToThread(transport.exchange, "HI", computer_id="34",
                               message_api="X.Y")

        def got_result(ignored):
            self.assertEqual(r.request.received_headers["x-computer-id"],
                             "34")
            self.assertEqual(r.request.received_headers["user-agent"],
                             "landscape-client/%s" % (VERSION,))
            self.assertEqual(r.request.received_headers["x-message-api"],
                             "X.Y")
            self.assertEqual(bpickle.loads(r.content), "HI")
        result.addCallback(got_result)
        return result

    def test_ssl_verification_negative(self):
        """
        If the SSL server provides a key which is not verified by the
        specified public key, then the client should immediately end
        the connection without uploading any message data.
        """
        self.log_helper.ignore_errors(PyCurlError)
        r = DataCollectingResource()
        context_factory = DefaultOpenSSLContextFactory(
            BADPRIVKEY, BADPUBKEY)
        port = reactor.listenSSL(0, server.Site(r), context_factory,
                                 interface="127.0.0.1")
        self.ports.append(port)
        transport = HTTPTransport(None, "https://localhost:%d/"
                                  % (port.getHost().port,), pubkey=PUBKEY)

        result = deferToThread(transport.exchange, "HI", computer_id="34",
                               message_api="X.Y")

        def got_result(ignored):
            self.assertIs(r.request, None)
            self.assertIs(r.content, None)
            self.assertTrue("server certificate verification failed"
                            in self.logfile.getvalue())
        result.addErrback(got_result)
        return result



from landscape.monitor.monitor import Monitor
from landscape.lib.persist import Persist
from landscape.tests.helpers import LandscapeTest, MonitorHelper
from landscape.broker.client import BrokerClientPlugin


class MonitorTest(LandscapeTest):

    helpers = [MonitorHelper]

    def test_persist(self):
        """
        A L{Monitor} instance has a C{persist} attribute.
        """
        self.monitor.persist.set("a", 1)
        self.assertEqual(self.monitor.persist.get("a"), 1)

    def test_flush_saves_persist(self):
        """
        The L{Monitor.flush} method saves any changes made to the persist
        database.
        """
        self.monitor.persist.set("a", 1)
        self.monitor.flush()

        persist = Persist()
        persist.load(self.monitor.persist_filename)
        self.assertEqual(persist.get("a"), 1)

    def test_flush_after_exchange(self):
        """
        The L{Monitor.exchange} method flushes the monitor after
        C{exchange} on all plugins has been called.
        """
        plugin = BrokerClientPlugin()
        plugin.exchange = lambda: self.monitor.persist.set("a", 1)
        self.monitor.add(plugin)
        self.monitor.exchange()

        persist = Persist()
        persist.load(self.monitor.persist_filename)
        self.assertEqual(persist.get("a"), 1)

    def test_flush_every_flush_interval(self):
        """
        The L{Monitor.flush} method gets called every C{flush_interval}
        seconds, and perists data to the disk.
        """
        self.monitor.persist.save = self.mocker.mock()
        self.monitor.persist.save(self.monitor.persist_filename)
        self.mocker.count(3)
        self.mocker.replay()
        self.reactor.advance(self.config.flush_interval * 3)

    def test_creating_loads_persist(self):
        """
        If C{persist_filename} exists, it is loaded by the constructor.
        """
        filename = self.makeFile()

        persist = Persist()
        persist.set("a", "Hi there!")
        persist.save(filename)

        monitor = Monitor(self.reactor, self.config, persist=Persist(),
                          persist_filename=filename)
        self.assertEqual(monitor.persist.get("a"), "Hi there!")



from landscape.sysinfo.sysinfo import SysInfoPluginRegistry
from landscape.sysinfo.memory import Memory
from landscape.tests.helpers import LandscapeTest


MEMINFO_SAMPLE = """
MemTotal:      2074536 kB
MemFree:        436468 kB
Buffers:        385596 kB
Cached:         672856 kB
SwapCached:          0 kB
Active:         708424 kB
Inactive:       705292 kB
HighTotal:     1178432 kB
HighFree:       137220 kB
LowTotal:       896104 kB
LowFree:        299248 kB
SwapTotal:     2562356 kB
SwapFree:      1562356 kB
Dirty:             300 kB
Writeback:           0 kB
AnonPages:      355388 kB
Mapped:         105028 kB
Slab:           152664 kB
SReclaimable:   136372 kB
SUnreclaim:      16292 kB
PageTables:       3124 kB
NFS_Unstable:        0 kB
Bounce:              0 kB
CommitLimit:   3599624 kB
Committed_AS:  1136296 kB
VmallocTotal:   114680 kB
VmallocUsed:     27796 kB
VmallocChunk:    86764 kB
"""


class MemoryTest(LandscapeTest):

    def setUp(self):
        super(MemoryTest, self).setUp()
        self.memory = Memory(self.makeFile(MEMINFO_SAMPLE))
        self.sysinfo = SysInfoPluginRegistry()
        self.sysinfo.add(self.memory)

    def test_run_returns_succeeded_deferred(self):
        self.assertIs(None, self.successResultOf(self.memory.run()))

    def test_run_adds_header(self):
        self.memory.run()
        self.assertEqual(self.sysinfo.get_headers(),
                         [("Memory usage", "27%"),
                          ("Swap usage", "39%")])



class UISettings(object):
    """
    A very thin wrapper around L{GSettings} to avoid having to know the
    L{BaseKey} and type information elsewhere.  In some future version it would
    be right to bind to change events here so we can react to people changing
    the settings in dconf, for now that is overkill.
    """

    BASE_KEY = "com.canonical.landscape-client-settings"

    def __init__(self, settings):
        self.settings = settings.new(self.BASE_KEY)

    def get_management_type(self):
        return self.settings.get_string("management-type")

    def set_management_type(self, value):
        self.settings.set_string("management-type", value)

    def get_computer_title(self):
        return self.settings.get_string("computer-title")

    def set_computer_title(self, value):
        self.settings.set_string("computer-title", value)

    def get_hosted_landscape_host(self):
        return self.settings.get_string("hosted-landscape-host")

    def get_hosted_account_name(self):
        return self.settings.get_string("hosted-account-name")

    def set_hosted_account_name(self, value):
        self.settings.set_string("hosted-account-name", value)

    def get_hosted_password(self):
        return self.settings.get_string("hosted-password")

    def set_hosted_password(self, value):
        self.settings.set_string("hosted-password", value)

    def get_local_landscape_host(self):
        return self.settings.get_string("local-landscape-host")

    def set_local_landscape_host(self, value):
        self.settings.set_string("local-landscape-host", value)

    def get_local_account_name(self):
        return self.settings.get_string("local-account-name")

    def set_local_account_name(self, value):
        self.settings.set_string("local-account-name", value)

    def get_local_password(self):
        return self.settings.get_string("local-password")

    def set_local_password(self, value):
        self.settings.set_string("local-password", value)
